# -*- coding: utf-8 -*-
"""The data model build on the Core Scientific Dataset Model."""
from __future__ import division
from __future__ import print_function

import json
from urllib.parse import urlparse

import numpy as np

from .csdm import as_dependent_variable  # lgtm [py/import-own-module] # NOQA
from .csdm import as_dimension  # lgtm [py/import-own-module] # NOQA
from .csdm import CSDM  # lgtm [py/import-own-module] # NOQA
from .csdm import DependentVariable  # lgtm [py/import-own-module] # NOQA
from .csdm import Dimension  # lgtm [py/import-own-module] # NOQA
from .csdm import LabeledDimension  # lgtm [py/import-own-module] # NOQA
from .csdm import LinearDimension  # lgtm [py/import-own-module] # NOQA
from .csdm import MonotonicDimension  # lgtm [py/import-own-module] # NOQA
from .dependent_variables import download  # lgtm [py/import-own-module] # NOQA
from .helper_functions import _preview  # lgtm [py/import-own-module] # NOQA
from .numpy_wrapper import apodize  # lgtm [py/import-own-module] # NOQA
from .tests import *  # lgtm [py/import-own-module] # NOQA
from .units import ScalarQuantity  # lgtm [py/import-own-module] # NOQA
from .units import string_to_quantity  # lgtm [py/import-own-module] # NOQA
from .utils import QuantityType  # lgtm [py/import-own-module] # NOQA
from .utils import validate  # lgtm [py/import-own-module] # NOQA

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__copyright__ = "Copyright 2019-2020, The CSDMpy Project."
__credits__ = ["Deepansh J. Srivastava"]
__license__ = "BSD License"
__maintainer__ = "Deepansh J. Srivastava"
__status__ = "Development"
__version__ = "0.3.4"

__all__ = [
    "parse_dict",
    "load",
    "loads",
    "new",
    "as_csdm",
    "as_dependent_variable",
    "as_dimension",
    "plot",
]


def _import_json(filename, verbose=False):
    res = urlparse(filename)
    if res[0] not in ["file", ""]:
        filename = download.download_file_from_url(filename, verbose)
    with open(filename, "rb") as f:
        content = f.read()
        return json.loads(str(content, encoding="UTF-8"))


def _check_csdm_root_key_value(dictionary):
    """Check the root level key and value type of the csdm object"""
    key_list_root = dictionary.keys()
    if "CSDM" in key_list_root:
        raise KeyError(
            "'CSDM' is not a valid keyword for the CSD model. Did you mean 'csdm'?"
        )

    if "csdm" not in key_list_root:
        raise KeyError("Missing a required `csdm` key from the data model.")

    # inside csdm object
    optional_keys = [
        "read_only",
        "timestamp",
        "geographic_coordinate",
        "application",
        "tags",
        "description",
    ]
    required_keys = ["version"]
    all_keys = optional_keys + required_keys

    key_list_csdm = list(dictionary["csdm"].keys())
    key_list_csdm_lower_case = [item.lower() for item in key_list_csdm]

    for i, item in enumerate(key_list_csdm):
        if item not in all_keys and key_list_csdm_lower_case[i] in all_keys:
            raise KeyError(
                f"{item} is an invalid key for the CSDM object. "
                f"Did you mean '{key_list_csdm_lower_case[i]}'?"
            )

    for item in required_keys:
        if item not in key_list_csdm:
            raise KeyError(f"Missing a required `{item}` key from the CSDM object.")

    _version = dictionary["csdm"]["version"]
    validate(_version, "version", str)


def parse_dict(dictionary):
    """Parse a CSDM compliant python dictionary and return a CSDM object.

    Args:
        dictionary: A CSDM compliant python dictionary.
    """
    optional_keys = [
        "read_only",
        "timestamp",
        "geographic_coordinate",
        "application",
        "tags",
        "description",
    ]

    _check_csdm_root_key_value(dictionary)

    _version = dictionary["csdm"]["version"]

    if "filename" in dictionary.keys():
        filename = dictionary["filename"]
    else:
        filename = None
    csdm = CSDM(filename=filename, version=_version)

    keys = dictionary["csdm"].keys()
    if "timestamp" in keys:
        _timestamp = dictionary["csdm"]["timestamp"]
        validate(_timestamp, "timestamp", str)
        csdm._timestamp = _timestamp

    if "dimensions" in keys:
        for dim in dictionary["csdm"]["dimensions"]:
            csdm.add_dimension(dim)

    if "dependent_variables" in keys:
        for dat in dictionary["csdm"]["dependent_variables"]:
            csdm.add_dependent_variable(dat)

    for key in optional_keys:
        if key in keys:
            setattr(csdm, "_" + key, dictionary["csdm"][key])

    return csdm


def load(filename=None, application=False, verbose=False):
    r"""
    Loads a .csdf/.csdfe file and returns an instance of the :ref:`csdm_api` class.

    The file must be a JSON serialization of the CSD Model.

    Example:
        >>> data1 = cp.load('local_address/file.csdf') # doctest: +SKIP
        >>> data2 = cp.load('url_address/file.csdf') # doctest: +SKIP

    Args:
        filename (str): A local or a remote address to the `.csdf or `.csdfe` file.
        application (bool): If true, the application metadata from application that
                last serialized the file will be imported. Default is False.
        verbose (bool): If the filename is a URL, this option will show the progress
                bar for the file download status, when True.

    Returns:
        A CSDM instance.
    """
    if filename is None:
        raise Exception("Missing the value for the required `filename` attribute.")

    dictionary = _import_json(filename, verbose)
    dictionary["filename"] = filename
    csdm_object = parse_dict(dictionary)

    if application is False:
        csdm_object.application = {}
        for dim in csdm_object.dimensions:
            dim.application = {}
            if hasattr(dim, "reciprocal") and dim.type != "label":
                dim.reciprocal.application = {}
        for dim in csdm_object.dependent_variables:
            dim.application = {}
            # if hasattr(dim., 'dimension indexes'):
            #     dim.reciprocal.application = {}
    # csdm_objects = []
    # for file_ in csdm_files:
    #     csdm_objects.append(_load(file_, application=application))
    return csdm_object


def loads(string):
    """
        Loads a JSON serialized string as a CSDM object.

        Args:
            string: A JSON serialized CSDM string.
        Returns:
            A CSDM object.

        Example:
            >>> object_from_string = cp.loads(cp.new('A test dump').dumps())
            >>> print(object_from_string.data_structure)  # doctest: +SKIP
            {
              "csdm": {
                "version": "1.0",
                "timestamp": "2019-10-21T20:33:17Z",
                "description": "A test dump",
                "dimensions": [],
                "dependent_variables": []
              }
            }
        """
    dictionary = json.loads(string)
    csdm_object = parse_dict(dictionary)
    return csdm_object


def new(description=""):
    r"""
    Creates a new instance of the :ref:`csdm_api` class containing a 0D{0} dataset.

    Args:
        description (str): A string describing the csdm object. This is optional.

    Example:
        >>> import csdmpy as cp
        >>> emptydata = cp.new(description='Testing Testing 1 2 3')
        >>> print(emptydata.data_structure)
        {
          "csdm": {
            "version": "1.0",
            "description": "Testing Testing 1 2 3",
            "dimensions": [],
            "dependent_variables": []
          }
        }

    Returns:
        A CSDM instance.
    """
    return CSDM(description=description)


def as_csdm(array, unit="", quantity_type="scalar"):
    """Generate and return a view of the nD numpy array as a csdm object.
    The nD array is the dependent variable of the csdm object of the given quantity
    type. The shape of the nD array is used to generate Dimension object of `linear`
    subtype.

    Args:
        array: The nD numpy array.
        unit: The unit for the dependent variable. The default is empty string.
        quantity_type: The quantity type of the dependent variable.

    Example:
        >>> array = np.arange(30).reshape(3, 10)
        >>> csdm_obj = cp.as_csdm(array)
        >>> print(csdm_obj)
        CSDM(
        DependentVariable(
        [[[ 0  1  2  3  4  5  6  7  8  9]
          [10 11 12 13 14 15 16 17 18 19]
          [20 21 22 23 24 25 26 27 28 29]]], quantity_type=scalar, numeric_type=int64),
        LinearDimension([0. 1. 2. 3. 4. 5. 6. 7. 8. 9.]),
        LinearDimension([0. 1. 2.])
        )
    """
    q_type = QuantityType(quantity_type)
    if q_type.p == 1:
        array = array[np.newaxis, :]

    if q_type.p != array.shape[0]:
        raise ValueError(
            f"Expecting exactly {q_type.p} components for quantity type, "
            f"`{quantity_type}`, found {array.shape[0]}. Make sure `array.shape[0]` "
            f"is equal to the number of components supported by {quantity_type}."
        )

    csdm = new()

    # dimensions should be added first
    shape = array.shape[::-1]
    for i in shape[:-1]:
        csdm.add_dimension(LinearDimension(count=i, increment="1"))

    csdm.add_dependent_variable(
        type="internal",
        components=array.ravel(),
        unit=unit,
        quantity_type=quantity_type,
    )

    return csdm


def plot(csdm_object, reverse_axis=None, range=None, **kwargs):
    """
    A supplementary function for plotting basic 1D and 2D datasets only.

    Args:
        csdm_object: The CSDM object.
        reverse_axis: An ordered array of boolean specifying which dimensions will be
            displayed on a reverse axis.
        range: A list of minimum and maxmim coordinates along the dimensions. The range
            along each dimension is given as [min, max]
        kwargs: Additional keyword arguments are used in matplotlib plotting functions.
            We implement the following matplotlib methods for the one and
            two-dimensional datasets.

            - The 1D{1} scalar dataset use the plt.plot() method.
            - The 1D{2} vector dataset use the plt.quiver() method.
            - The 2D{1} scalar dataset use the plt.imshow() method if the two
              dimensions have a `linear` subtype. If any one of the dimension is
              `monotonic`, plt.NonUniformImage() method is used instead.
            - The 2D{2} vector dataset use the plt.quiver() method.
            - The 2D{3} pixel dataset use the plt.imshow(), assuming the pixel dataset
              as an RGB image.

    Returns:
        A matplotlib figure instance.

    Example:
        >>> cp.plot(data_object) # doctest: +SKIP
    """
    return _preview(csdm_object, reverse_axis, range, **kwargs)



# -*- coding: utf-8 -*-
"""The Dimension object: attributes and methods."""
import warnings
from copy import deepcopy

import numpy as np

from .labeled import LabeledDimension  # lgtm [py/import-own-module]
from .linear import LinearDimension  # lgtm [py/import-own-module]
from .monotonic import MonotonicDimension  # lgtm [py/import-own-module]
from csdmpy.units import string_to_quantity  # lgtm [py/import-own-module]
from csdmpy.utils import _get_dictionary  # lgtm [py/import-own-module]
from csdmpy.utils import validate  # lgtm [py/import-own-module]

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["Dimension"]


functional_dimension = ["linear"]


class Dimension:
    r"""
    Create an instance of the Dimension class.

    An instance of this class describes a dimension of a multi-dimensional system.
    In version 1.0 of the CSD model, there are three subtypes of the Dimension class:

    - :ref:`linearDimension_uml`,
    - :ref:`monotonicDimension_uml`, and
    - :ref:`labeledDimension_uml`.

    **Creating an instance of a dimension object**

    There are two ways of creating a new instance of a Dimension class.

    *From a python dictionary containing valid keywords.*

    .. doctest::

        >>> from csdmpy import Dimension
        >>> dimension_dictionary = {
        ...     'type': 'linear',
        ...     'description': 'test',
        ...     'increment': '5 G',
        ...     'count': 10,
        ...     'coordinates_offset': '10 mT',
        ...     'origin_offset': '10 T'
        ... }
        >>> x = Dimension(dimension_dictionary)

    Here, `dimension_dictionary` is the python dictionary.

    *From valid keyword arguments.*

    .. doctest::

        >>> x = Dimension(type = 'linear',
        ...               description = 'test',
        ...               increment = '5 G',
        ...               count = 10,
        ...               coordinates_offset = '10 mT',
        ...               origin_offset = '10 T')

    """

    __slots__ = ("subtype",)

    _immutable_objects_ = ()

    def __init__(self, *args, **kwargs):
        """Initialize an instance of Dimension object."""
        default = {
            "type": None,  # valid for all dimension subtypes
            "description": "",  # valid for all dimension subtypes
            "count": None,  # valid for linear subtype
            "increment": None,  # valid for linear subtype
            "labels": None,  # valid for labeled subtype
            "coordinates": None,  # valid for monotonic subtype
            "coordinates_offset": None,  # valid for linear subtype
            "origin_offset": None,  # valid for linear subtype
            "complex_fft": False,  # valid for linear subtype
            "period": None,  # valid for monotonic and linear subtypes
            "quantity_name": None,  # valid for monotonic and linear subtypes
            "label": "",  # valid for all dimension subtypes
            "application": {},  # valid for all dimension subtypes
            "reciprocal": {  # valid for monotonic and linear subtypes
                "increment": None,  # valid for monotonic and linear subtypes
                "coordinates_offset": None,  # valid for monotonic and linear subtypes
                "origin_offset": None,  # valid for monotonic and linear subtypes
                "period": None,  # valid for monotonic and linear subtypes
                "quantity_name": None,  # valid for monotonic and linear subtypes
                "label": "",  # valid for monotonic and linear subtypes
                "description": "",  # valid for monotonic and linear subtypes
                "application": {},  # valid for monotonic and linear subtypes
            },
        }

        default_keys = default.keys()
        input_dict = _get_dictionary(*args, **kwargs)
        input_keys = input_dict.keys()

        if "type" not in input_keys:
            raise KeyError("Missing a required 'type' key from the Dimension object.")

        if "reciprocal" in input_keys:
            input_subkeys = input_dict["reciprocal"].keys()
        for key in input_keys:
            if key in default_keys:
                if key == "reciprocal":
                    for subkey in input_subkeys:
                        default[key][subkey] = input_dict[key][subkey]
                else:
                    default[key] = input_dict[key]

        self.__validate_key_value__(default)

        if default["type"] == "labeled":
            self.subtype = LabeledDimension(**default)

        if default["type"] == "monotonic":
            self.subtype = MonotonicDimension(values=default["coordinates"], **default)

        if default["type"] == "linear":
            self.subtype = self._linear(default)

    @staticmethod
    def __validate_key_value__(default):
        _valid_types = ["monotonic", "linear", "labeled"]

        type_ = default["type"]
        message = (
            f"The value, '{type_}', is invalid for the `type` attribute of the "
            "Dimension object. The allowed values are 'monotonic', 'linear' and "
            "'labeled'."
        )

        if default["type"] not in _valid_types:
            raise ValueError(message)

        if default["type"] == "labeled" and default["labels"] is None:
            raise KeyError(
                "Missing a required `labels` key from the LabeledDimension object."
            )

        if default["type"] == "monotonic" and default["coordinates"] is None:
            raise KeyError(
                "Missing a required `coordinates` key from the MonotonicDimension "
                "object."
            )

    def _linear(self, default):
        """Create and assign a linear dimension."""
        missing_key = ["increment", "count"]

        for item in missing_key:
            if default[item] is None:
                raise KeyError(
                    f"Missing a required `{item}` key from the LinearDimension object."
                )

        validate(default["count"], "count", int)

        return LinearDimension(**default)

    def __repr__(self):
        """String representation of object."""
        return self.subtype.__repr__()

    def __str__(self):
        """String representation of object."""
        return self.subtype.__str__()

    def __eq__(self, other):
        """Overrides the default implementation."""
        if isinstance(other, Dimension):
            if self.subtype == other.subtype:
                return True
            return False
        if isinstance(other, (LinearDimension, LabeledDimension, MonotonicDimension)):
            if self.subtype == other:
                return True
            return False
        return False

    def __mul__(self, other):
        """Multiply the Dimension object by a scalar."""
        return self.subtype.__mul__(other)

    def __imul__(self, other):
        """Multiply the Dimension object by a scalar, in-place."""
        return self.subtype.__imul__(other)

    def __truediv__(self, other):
        """Divide the Dimension object by a scalar."""
        return self.subtype.__truediv__(other)

    def __itruediv__(self, other):
        """Divide the Dimension object by a scalar, in-place."""
        return self.subtype.__itruediv__(other)

    # ======================================================================= #
    #                          Dimension Attributes                           #
    # ======================================================================= #
    @property
    def absolute_coordinates(self):
        r"""
        Absolute coordinates, :math:`\bf X_k^{\rm{abs}}`, along the dimension.

        This attribute is only *valid* for quantitative dimensions, that is,
        `linear` and `monotonic` dimensions. The absolute coordinates are given as

        .. math::

            \mathbf{X}_k^\mathrm{abs} = \mathbf{X}_k + o_k \mathbf{1}

        where :math:`\mathbf{X}_k` are the coordinates along the dimension and
        :math:`o_k` is the :attr:`~csdmpy.Dimension.origin_offset`.
        For example, consider

        .. doctest::

            >>> print(x.origin_offset)
            10.0 T
            >>> print(x.coordinates[:5])
            [100. 105. 110. 115. 120.] G

        then the absolute coordinates are

        .. doctest::

            >>> print(x.absolute_coordinates[:5])
            [100100. 100105. 100110. 100115. 100120.] G

        For `linear` dimensions, the order of the `absolute_coordinates`
        further depend on the value of the
        :attr:`~csdmpy.Dimension.complex_fft` attributes. For
        examples, when the value of the `complex_fft` attribute is True,
        the absolute coordinates are

        .. doctest::

            >>> x.complex_fft = True
            >>> print(x.absolute_coordinates[:5])
            [100075. 100080. 100085. 100090. 100095.] G

        .. testsetup::

            >>> x.complex_fft = False

        Returns:
            A Quantity array of absolute coordinates for quantitative dimensions, `i.e`
            `linear` and `monotonic`.

        Raises:
            AttributeError: For labeled dimensions.
        """
        return self.subtype.absolute_coordinates

    @property
    def application(self):
        """
        Application metadata dictionary of the dimension object.

        .. doctest::

            >>> print(x.application)
            {}

        The application attribute is where an application can place its own
        metadata as a python dictionary object containing application specific
        metadata, using a reverse domain name notation string as the attribute
        key, for example,

        .. doctest::

            >>> x.application = {
            ...     "com.example.myApp" : {
            ...         "myApp_key": "myApp_metadata"
            ...      }
            ... }
            >>> print(x.application)
            {'com.example.myApp': {'myApp_key': 'myApp_metadata'}}

        Returns:
            A python dictionary containing dimension application metadata.
        """
        return self.subtype.application

    @application.setter
    def application(self, value):
        self.subtype.application = value

    @property
    def axis_label(self):
        r"""
        Formatted string for displaying label along the dimension axis.

        This attribute is not a part of the original core scientific dataset
        model, however, it is a convenient supplementary attribute that provides
        a formated string ready for labeling dimension axes.
        For quantitative dimensions, this attributes returns a string,
        `label / unit`,  if the `label` is a non-empty string, otherwise,
        `quantity_name / unit`. Here
        :attr:`~csdmpy.Dimension.quantity_name` and
        :attr:`~csdmpy.Dimension.label` are the attributes of the
        :ref:`dim_api` instances, and `unit` is the unit associated with the
        coordinates along the dimension. For examples,

        .. doctest::

            >>> x.label
            'field strength'
            >>> x.axis_label
            'field strength / (G)'

        For `labeled` dimensions, this attribute returns `label`.

        Returns:
            A formated string of label.

        Raises:
            AttributeError: When assigned a value.
        """
        return self.subtype.axis_label

    @property
    def coordinates(self):
        r"""
        Coordinates, :math:`{\bf X}_k`, along the dimension.

        Example:
            >>> print(x.coordinates)
            [100. 105. 110. 115. 120. 125. 130. 135. 140. 145.] G

        For `linear` dimensions, the order of the `coordinates` also depend on the
        value of the :attr:`~csdmpy.Dimension.complex_fft` attributes.
        For examples, when the value of the `complex_fft` attribute is True,
        the coordinates are

        .. doctest::

            >>> x.complex_fft = True
            >>> print(x.coordinates)
            [ 75.  80.  85.  90.  95. 100. 105. 110. 115. 120.] G

        .. testsetup::

            >>> x.complex_fft = False

        Returns:
            A Quantity array of coordinates for quantitative dimensions, `i.e.` `linear`
            and `monotonic`.

        Returns:
            A Numpy array for labeled dimensions.

        Raises:
            AttributeError: For dimensions with subtype `linear`.
        """
        return self.subtype.coordinates

    @coordinates.setter
    def coordinates(self, value):
        self.subtype.coordinates = value

    @property
    def coords(self):
        """Alias for the `coordinates` attribute."""
        return self.coordinates

    @coords.setter
    def coords(self, value):
        self.coordinates = value

    @property
    def data_structure(self):
        r"""
        Json serialized string describing the Dimension class instance.

        This supplementary attribute is useful for a quick preview of the dimension
        object. The attribute cannot be modified.

        .. doctest::

            >>> print(x.data_structure)
            {
              "type": "linear",
              "description": "This is a test",
              "count": 10,
              "increment": "5.0 G",
              "coordinates_offset": "10.0 mT",
              "origin_offset": "10.0 T",
              "quantity_name": "magnetic flux density",
              "label": "field strength"
            }

        Returns:
            A json serialized string of the dimension object.
        Raises:
            AttributeError: When modified.
        """
        return self.subtype.data_structure

    @property
    def description(self):
        """
        Brief description of the dimension object.

        The default value is an empty string, ''. The attribute may be
        modified, for example,

        .. doctest::

            >>> print(x.description)
            This is a test

            >>> x.description = 'This is a test dimension.'

        Returns:
            A string of UTF-8 allows characters describing the dimension.

        Raises:
            TypeError: When the assigned value is not a string.
        """
        return self.subtype.description

    @description.setter
    def description(self, value):
        self.subtype.description = value

    @property
    def complex_fft(self):
        r"""
        If true, the coordinates are the ordered as the output of a complex fft.

        This attribute is only `valid` for the Dimension instances with `linear`
        subtype.
        The value of this attribute is a boolean specifying if the coordinates along
        the dimension are evaluated as the output of a complex fast Fourier transform
        (FFT) routine.
        For example, consider the following Dimension object,

        .. doctest::

            >>> test = Dimension(
            ...            type='linear',
            ...	           increment = '1',
            ...            count = 10
            ...        )

            >>> test.complex_fft
            False
            >>> print(test.coordinates)
            [0. 1. 2. 3. 4. 5. 6. 7. 8. 9.]

            >>> test.complex_fft = True
            >>> print(test.coordinates)
            [-5. -4. -3. -2. -1.  0.  1.  2.  3.  4.]

        Returns:
            A Boolean.

        Raises:
            TypeError: When the assigned value is not a boolean.
        """
        return self.subtype.complex_fft

    @complex_fft.setter
    def complex_fft(self, value):
        self.subtype.complex_fft = value

    @property
    def increment(self):
        r"""
        Increment along a `linear` dimension.

        The attribute is only `valid` for Dimension instances with the subtype
        `linear`. When assigning a value, the dimensionality of the value must
        be consistent with the dimensionality of other members specifying the
        dimension.

        Example:
            >>> print(x.increment)
            5.0 G
            >>> x.increment = "0.1 G"
            >>> print(x.coordinates)
            [100.  100.1 100.2 100.3 100.4 100.5 100.6 100.7 100.8 100.9] G

        Returns:
            A Quantity instance with the increment along the dimension.

        Raises:
            AttributeError: For dimension with subtypes other than `linear`.
            TypeError: When the assigned value is not a string containing a quantity
                       or a Quantity object.
        """
        # .. note:: The sampling interval along a grid dimension and the
        #     respective reciprocal grid dimension follow the Nyquist–Shannon
        #     sampling theorem. Therefore, updating the ``increment``
        #     will automatically trigger an update on its reciprocal
        #     counterpart.
        return self.subtype.increment

    @increment.setter
    def increment(self, value):
        self.subtype.increment = value

    @property
    def coordinates_offset(self):
        r"""
        Offset corresponding to the zero of the indexes array, :math:`\mathbf{J}_k`.

        When assigning a value, the dimensionality of the value must be consistent with
        the dimensionality of the other members specifying the dimension.

        Example:
            >>> print(x.coordinates_offset)
            10.0 mT
            >>> x.coordinates_offset = "0 T"
            >>> print(x.coordinates)
            [ 0.  5. 10. 15. 20. 25. 30. 35. 40. 45.] G

        The attribute is `invalid` for `labeled` dimensions.

        Returns:
            A Quantity instance with the coordinates offset.

        Raises:
            AttributeError: For `labeled` dimensions.
            TypeError: When the assigned value is not a string containing a quantity
                       or a Quantity object.
        """
        return self.subtype.coordinates_offset

    @coordinates_offset.setter
    def coordinates_offset(self, value):
        self.subtype.coordinates_offset = value

    @property
    def label(self):
        r"""
        Label associated with the dimension.

        Example:
            >>> print(x.label)
            field strength
            >>> x.label = 'magnetic field strength'

        Returns:
            A string containing the label.
        Raises:
            TypeError: When the assigned value is not a string.
        """
        return self.subtype.label

    @label.setter
    def label(self, label=""):
        self.subtype.label = label

    @property
    def count(self):
        r"""
        Number of coordinates, :math:`N_k \ge 1`, along the dimension.

        Example:
            >>> print(x.count)
            10
            >>> x.count = 5

        Returns:
            An Integer specifying the number of coordinates along the dimension.

        Raises:
            TypeError: When the assigned value is not an integer.
        """
        return self.subtype._count

    @count.setter
    def count(self, value):
        self.subtype.count = value

    @property
    def origin_offset(self):
        r"""
        Origin offset, :math:`o_k`, along the dimension.

        When assigning a value, the dimensionality of the value must be consistent
        with the dimensionality of other members specifying the dimension.

        Example:
            >>> print(x.origin_offset)
            10.0 T
            >>> x.origin_offset = "1e5 G"

        The origin offset only affect the absolute_coordinates along the dimension.
        This attribute is `invalid` for `labeled` dimensions.

        Returns:
            A Quantity instance with the origin offset.

        Raises:
            AttributeError: For `labeled` dimensions.
            TypeError: When the assigned value is not a string containing a quantity
                       or a Quantity object.
        """
        return self.subtype.origin_offset

    @origin_offset.setter
    def origin_offset(self, value):
        self.subtype.origin_offset = value

    @property
    def period(self):
        r"""
        Period of the dimension.

        The default value of the period is infinity, i.e., the dimension is
        non-periodic.

        Example:
            >>> print(x.period)
            inf G
            >>> x.period = '1 T'

        To assign a dimension as non-periodic, one of the following may be
        used,

        .. doctest::

            >>> x.period = '1/0 T'
            >>> x.period = 'infinity µT'
            >>> x.period = '∞ G'

        .. Attention::
            The physical quantity of the period must be consistent with other
            physical quantities specifying the dimension.

        Returns:
            A Quantity instance with the period of the dimension.

        Raises:
            AttributeError: For `labeled` dimensions.
            TypeError: When the assigned value is not a string containing a quantity
                       or a Quantity object.
        """
        return self.subtype.period

    @period.setter
    def period(self, value=None):
        self.subtype.period = value

    @property
    def quantity_name(self):
        r"""
        Quantity name associated with the physical quantities specifying the dimension.

        The attribute is `invalid` for the labeled dimension.

        .. doctest::

            >>> print(x.quantity_name)
            magnetic flux density

        Returns:
            A string with the `quantity name`.

        Raises:
            AttributeError: For `labeled` dimensions.
            NotImplementedError: When assigning a value.
        """
        return self.subtype.quantity_name

    @quantity_name.setter
    def quantity_name(self, value):
        self.subtype.quantity_name = value

    @property
    def type(self):
        r"""
        The dimension subtype.

        There are three *valid* subtypes of Dimension class. The valid
        literals are given by the :ref:`dimObjectSubtype_uml` enumeration.

        .. doctest::

            >>> print(x.type)
            linear

        Returns:
            A string with a valid dimension subtype.

        Raises:
            AttributeError: When the attribute is modified.
        """
        return self.subtype.type

    @property
    def labels(self):
        r"""
        Ordered list of labels along the `Labeled` dimension.

        Consider the following labeled dimension,

        .. doctest::

            >>> x2 = Dimension(
            ...         type='labeled',
            ...         labels=['Cu', 'Ag', 'Au']
            ...      )

        then the lables along the labeled dimension are

        .. doctest::

            >>> print(x2.labels)
            ['Cu' 'Ag' 'Au']

        .. note::
            For Labeled dimension, the :attr:`~csdmpy.Dimension.coordinates`
            attribute is an alias of :attr:`~csdmpy.Dimension.labels`
            attribute. For example,

            .. doctest::

                >>> np.all(x2.coordinates == x2.labels)
                True

        In the above example, ``x2`` is an instance of the :ref:`dim_api` class with
        `labeled` subtype.

        Returns:
             A Numpy array with labels along the dimension.

        Raises:
            AttributeError: For dimensions with subtype other than `labeled`.
        """
        return self.coordinates

    @labels.setter
    def labels(self, array):
        self.subtype.labels = array

    @property
    def reciprocal(self):
        r"""
        An instance of the ReciprocalDimension class.

        The attributes of ReciprocalDimension class are:
            - coordinates_offset
            - origin_offset
            - period
            - quantity_name
            - label
        where the definition of each attribute is the same as the corresponding
        attribute from the Dimension instance.
        """
        return self.subtype.reciprocal

    # ======================================================================= #
    #                           Dimension Methods                             #
    # ======================================================================= #

    def _copy_metadata(self, obj, copy=False):
        """Copy Dimension metadata"""
        self.subtype._type = obj.subtype._type
        self.subtype._copy_metadata(obj)

    def to_dict(self):
        """Alias to the `dict()` method of the class."""
        return self.dict()

    def dict(self):
        r"""
        Return Dimension object as a python dictionary.

        Example:
            >>> x.dict() # doctest: +SKIP
            {'type': 'linear', 'description': 'This is a test', 'count': 10,
            'increment': '5.0 G', 'coordinates_offset': '10.0 mT',
            'origin_offset': '10.0 T', 'quantity_name': 'magnetic flux density',
            'label': 'field strength'}
        """
        return self.subtype.dict()

    def is_quantitative(self):
        r"""
        Return True if the dependent variable is quantitative.

        Example:
            >>> x.is_quantitative()
            True
        """
        return self.subtype.is_quantitative()

    def to(self, unit="", equivalencies=None):
        r"""
        Convert the coordinates along the dimension to the unit, `unit`.

        This method is a wrapper of the `to` method from the
        `Quantity <http://docs.astropy.org/en/stable/api/\
        astropy.units.Quantity.html#astropy.units.Quantity.to>`_ class
        and is only `valid` for physical dimensions.

        Example:
            >>> print(x.coordinates)
            [100. 105. 110. 115. 120. 125. 130. 135. 140. 145.] G
            >>> x.to('mT')
            >>> print(x.coordinates)
            [10.  10.5 11.  11.5 12.  12.5 13.  13.5 14.  14.5] mT

        Args:
            `unit` : A string containing a unit with the same dimensionality as the
                     coordinates along the dimension.

        Raises:
            AttributeError: For `labeled` dimensions.
        """
        self.subtype.to(unit, equivalencies)

    def copy(self):
        """Return a copy of the Dimension object."""
        return deepcopy(self)

    def reciprocal_coordinates(self):
        """Return reciprocal coordinates assuming Nyquist-shannan theorem."""
        return self.subtype.reciprocal_coordinates()

    def reciprocal_increment(self):
        """Return reciprocal increment assuming Nyquist-shannan theorem."""
        return self.subtype.reciprocal_coordinates()


def as_dimension(array, unit="", type=None, **kwargs):
    """Generate and return a Dimension object from a 1D numpy array.

    Args:
        array: A 1D numpy array.
        unit: The unit of the coordinates along the dimension.
        type: The dimension type. Valid values are linear, monotonic, labeled, or
                None. If the value is None, let us decide. The default value is None.
        kwargs: Additional keyword arguments from the Dimension class.

    Example:
        >>> array = np.arange(15)*0.5
        >>> dim_object = cp.as_dimension(array)
        >>> print(dim_object)
        LinearDimension([0.  0.5 1.  1.5 2.  2.5 3.  3.5 4.  4.5 5.  5.5 6.  6.5 7. ])

        >>> array = ['The', 'great', 'circle']
        >>> dim_object = cp.as_dimension(array, label='in the sky')
        >>> print(dim_object)
        LabeledDimension(['The' 'great' 'circle'])
    """
    array = __check_array_for_dimension__(array, type)

    if type is None:
        return _generic_dimensions(array, unit, **kwargs)

    if type == "linear":
        obj = _linear_dimension(array, unit, **kwargs)
        if obj is not None:
            return obj
        raise ValueError("Invalid array for LinearDimension object.")

    if type == "monotonic":
        obj = _monotonic_dimension(array, unit, **kwargs)
        if obj is not None:
            return obj
        raise ValueError("Invalid array for MonotonicDimension object.")

    if type == "labeled":
        if unit != "":
            warnings.warn("Ignoring unit argument for LabeledDimension object.")
        return LabeledDimension(labels=array.tolist(), **kwargs)


def __check_array_for_dimension__(array, type):
    options = [None, "linear", "monotonic", "labeled"]
    if type not in options:
        raise ValueError(f"Invalid value for `type`. Allowed values are {options}.")

    if not isinstance(array, (list, np.ndarray)):
        raise ValueError(
            f"Cannot convert {array.__class__.__name__} to a Dimension object."
        )

    array = np.asarray(array)
    if array.ndim != 1:
        raise ValueError(
            f"Cannot convert a {array.ndim} dimensional array to a Dimension object."
        )
    return array


def _generic_dimensions(array, unit, className="Dimension", **kwargs):
    """Return a dimension object based on the array coordinates."""
    # labeled
    if str(array.dtype)[:2] in [">U", "<U"]:
        if unit != "":
            warnings.warn("Ignoring unit argument for LabeledDimension.")
        return LabeledDimension(labels=array.tolist(), **kwargs)

    # linear
    obj = _linear_dimension(array, unit, className, **kwargs)
    if obj is not None:
        return obj

    # monotonic
    obj = _monotonic_dimension(array, unit, **kwargs)
    if obj is not None:
        return obj

    raise ValueError("Invalid array for Dimension object.")


def _linear_dimension(array, unit, className="LinearDimension", **kwargs):
    """Return a LinearDimension is array is linear, else None."""
    increment = array[1] - array[0]
    if increment == 0:
        raise ValueError(f"Invalid array for {className} object.")

    if np.allclose(np.diff(array, 1), increment):

        if str(unit) != "":
            unit = f"({unit})"
        return LinearDimension(
            count=array.size,
            increment=f"{increment} {unit}",
            coordinates_offset=f"{array[0]} {unit}",
            **kwargs,
        )


def _monotonic_dimension(array, unit, **kwargs):
    """Return a MonotonicDimension is array is monotonic, else None."""
    if np.all(np.diff(array, 1) > 0) or np.all(np.diff(array, 1) < 0):
        return MonotonicDimension(
            coordinates=array * string_to_quantity(unit), **kwargs
        )



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from csdmpy.dependent_variables import check_sparse_sampling_key_value
from csdmpy.dependent_variables.sparse import SparseSampling

sparse_sampling = {
    "dimension_indexes": [0],
    "sparse_grid_vertexes": [0, 5, 10, 15, 20, 25],
}


def test_01():
    sp = SparseSampling(**sparse_sampling)
    assert sp.dimension_indexes == [0]

    assert np.all(sp.sparse_grid_vertexes == [0, 5, 10, 15, 20, 25])

    assert sp.description == ""
    assert sp.encoding == "none"
    assert sp.unsigned_integer_type.value == "int64"
    assert sp.application == {}


def test_02():
    sparse_sampling = {
        "dimension_indexes": [0],
    }

    error = "Missing a required `sparse_grid_vertexes`"
    with pytest.raises(KeyError, match=".*{0}.*".format(error)):
        check_sparse_sampling_key_value({"sparse_sampling": sparse_sampling})


def test_03():
    sparse_sampling = {
        "sparse_grid_vertexes": [0, 5, 10, 15, 20, 25],
    }

    error = "Missing a required `dimension_indexes`"
    with pytest.raises(KeyError, match=".*{0}.*".format(error)):
        check_sparse_sampling_key_value({"sparse_sampling": sparse_sampling})


def test_04():
    sparse_sampling = {
        "encoding": "base64",
        "dimension_indexes": [0],
        "sparse_grid_vertexes": [0, 5, 10, 15, 20, 25],
    }
    error = "Missing a required `unsigned_integer_type`"
    with pytest.raises(KeyError, match=".*{0}.*".format(error)):
        check_sparse_sampling_key_value({"sparse_sampling": sparse_sampling})


def test_05():
    sparse_sampling = {
        "encoding": "base64",
        "unsigned_integer_type": "float32",
        "dimension_indexes": [0],
        "sparse_grid_vertexes": [0, 5, 10, 15, 20, 25],
    }
    error = "float32 is an invalid `unsigned_integer_type` enumeration"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        check_sparse_sampling_key_value({"sparse_sampling": sparse_sampling})


def test_06():
    sp1 = SparseSampling(**sparse_sampling)
    sp2 = SparseSampling(**sparse_sampling)

    assert sp1 == sp2



# -*- coding: utf-8 -*-
"""The SparseSampling class."""
from copy import deepcopy

import numpy as np

from csdmpy.dependent_variables.decoder import Decoder
from csdmpy.utils import check_encoding
from csdmpy.utils import NumericType
from csdmpy.utils import QuantityType
from csdmpy.utils import validate

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["SparseSampling"]


class SparseSampling:
    r"""Declare a SparseSampling class."""

    __slots__ = (
        "_sparse_dimensions_indexes",
        "_sparse_grid_vertexes",
        "_encoding",
        "_quantity_type",
        "_unsigned_integer_type",
        "_description",
        "_application",
    )

    def __init__(
        self,
        dimension_indexes,
        sparse_grid_vertexes,
        encoding="none",
        quantity_type="scalar",
        unsigned_integer_type="int64",
        description="",
        application={},
        **kwargs,
    ):
        """Initialize a SparseDimension class."""
        self.encoding = encoding
        self._unsigned_integer_type = NumericType(unsigned_integer_type)
        self._quantity_type = QuantityType(quantity_type)
        self.description = description
        self.application = application
        self._sparse_dimensions_indexes = dimension_indexes
        self._sparse_grid_vertexes = Decoder(
            self._encoding,
            self._quantity_type,
            [sparse_grid_vertexes],
            self._unsigned_integer_type.dtype,
        )

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, SparseSampling):
            check = [
                np.all(
                    self._sparse_dimensions_indexes == other._sparse_dimensions_indexes
                ),
                np.all(self._sparse_grid_vertexes == other._sparse_grid_vertexes),
                self._encoding == other._encoding,
                self._quantity_type == other._quantity_type,
                self._unsigned_integer_type == other._unsigned_integer_type,
                self._description == other._description,
                self._application == other._application,
            ]
            if False in check:
                return False
            return True
        return False

    # ----------------------------------------------------------------------- #
    #                                 Attributes                              #
    # ----------------------------------------------------------------------- #

    @property
    def encoding(self):
        r"""Return the data encoding method."""
        return deepcopy(self._encoding)

    @encoding.setter
    def encoding(self, value):
        self._encoding = validate(value, "encoding", str, check_encoding)

    @property
    def unsigned_integer_type(self):
        r"""Return the unsigned integer type of data values."""
        return deepcopy(self._unsigned_integer_type)

    @unsigned_integer_type.setter
    def unsigned_integer_type(self, value):
        self._unsigned_integer_type.update(value)

    @property
    def application(self):
        """Return an application metadata dictionary."""
        return deepcopy(self._application)

    @application.setter
    def application(self, value):
        self._application = validate(value, "application", dict)

    @property
    def description(self):
        r"""Return the description of the object."""
        return deepcopy(self._description)

    @description.setter
    def description(self, value):
        self._description = validate(value, "description", str)

    @property
    def dimension_indexes(self):
        """List of dimension indexes corresponding to sparse dimensions."""
        return deepcopy(self._sparse_dimensions_indexes)

    @property
    def sparse_grid_vertexes(self):
        """List of grid vertexes corresponding to sparse dimensions."""
        return deepcopy(self._sparse_grid_vertexes)



# -*- coding: utf-8 -*-
"""Decoder for components' encoding types."""
import base64

import numpy as np


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["Decoder"]


class Decoder:
    def __new__(self, encoding, quantity_type, components, dtype):
        """
        Decode the components based on the encoding key value.

        The valid encodings are 'base64', 'none' (text), and 'raw' (binary).
        """

        if encoding != "raw":
            check_number_of_components_and_encoding_type(len(components), quantity_type)
        component_len = quantity_type.p
        method = getattr(self, "decode_" + encoding)
        return method(components, dtype, component_len)

    @staticmethod
    def decode_base64(components, dtype, component_len=None):
        components = np.asarray(
            [np.frombuffer(base64.b64decode(item), dtype=dtype) for item in components]
        )
        return components

    @staticmethod
    def decode_none(components, dtype, component_len=None):
        if dtype in ["<c8", "<c16", ">c8", ">c16"]:
            components = np.asarray(
                [
                    np.asarray(item[0::2]) + 1j * np.asarray(item[1::2])
                    for item in components
                ],
                dtype=dtype,
            )
        else:
            # components = np.asarray(
            #     [np.asarray(item) for item in components], dtype=dtype
            # )
            components = np.asarray(components, dtype=dtype)
        return components

    @staticmethod
    def decode_raw(components, dtype, component_len=None):
        components = np.frombuffer(components, dtype=dtype)
        components.shape = component_len, int(components.size / component_len)
        return components


def check_number_of_components_and_encoding_type(length, quantity_type):
    """Verify the consistency of encoding wrt the number of components."""
    if length != quantity_type.p:
        raise Exception(
            f"The quantity_type, '{quantity_type.value}', requires exactly "
            f"{quantity_type.p} component(s), found {length} components."
        )



# -*- coding: utf-8 -*-
"""
A wrapper to the Numpy functions used to apodize the components along a given
dimension. If :math:`x` is  the coordinates along the dimension, the apodization
function follows,

.. math::
        y = function(a x),

where :math:`a` is the argument of the function. The dimensionality of :math:`a`
must be reciprocal of that of :math:`x`.

The list of supported functions are:
    - sin
    - cos
    - ...
"""
import numpy as np

from csdmpy.csdm import _get_new_csdm_object_after_apodization

__all__ = ("sin", "cos", "tan", "arcsin", "arccos", "arctan", "exp")


# Trigonometric functions


def sin(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\sin(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the sine of the
                   dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.sin, arg, dimension)


def cos(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\cos(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the cosine of the
                   dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.cos, arg, dimension)


def tan(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\tan(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the tangent of the
                   dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.tan, arg, dimension)


def arcsin(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\arcsin(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the inverse sine of the
                   dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.arcsin, arg, dimension)


def arccos(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\arccos(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the inverse cosine of
                   the dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.arccos, arg, dimension)


def arctan(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\arctan(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the inverse tangent of
                   the dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.arctan, arg, dimension)


def exp(csdm, arg, dimension=0):
    r"""
    Apodize the components along the `dimension` with :math:`\exp(a x)`.

    Args:
        csdm: A CSDM object.
        arg: String or Quantity object. The function argument :math:`a`.
        dimension: An integer or tuple of `m` integers cooresponding to the
                   index/indices of the dimensions along which the exp of the
                   dependent variable components is performed.
    Return:
        A CSDM object with `d-m` dimensions, where `d` is the total
        number of dimensions from the original `csdm` object.
    """
    return _get_new_csdm_object_after_apodization(csdm, np.exp, arg, dimension)



# -*- coding: utf-8 -*-
import numpy as np

import csdmpy as cp
from csdmpy.utils import _check_dimension_indices  # lgtm [py/import-own-module]
from csdmpy.utils import _get_broadcast_shape  # lgtm [py/import-own-module]


def fft(csdm, axis=0):
    """Perform a FFT along the given `dimension=axis`."""
    index = _check_dimension_indices(len(csdm.dimensions), axis)[0]
    # for index in indexes:
    if csdm.dimensions[axis].type != "linear":
        raise NotImplementedError(
            f"The FFT method is not available for Dimension objects with "
            f"subtype, {csdm.dimensions[axis].type}."
        )

    csdm_new = csdm.astype(np.complex128)
    dimension_object = csdm_new.dimensions[axis]
    if not isinstance(dimension_object, cp.LinearDimension):
        dimension_object = dimension_object.subtype

    unit_in = dimension_object._unit
    # the following coordinates does not include the coordinates offset and have are
    # given in the vase unit of the dimension.
    coordinates = dimension_object._coordinates.to(unit_in).value
    coordinates_offset = dimension_object._coordinates_offset.to(unit_in).value

    unit_out = (1 / unit_in).unit
    # the following coordinates does not include the coordinates offset and have are
    # given in the vase unit of the dimension.
    coordinates_res = dimension_object._reciprocal_coordinates().to(unit_out).value
    coordinates_offset_res = dimension_object.reciprocal._coordinates_offset
    coordinates_offset_res = coordinates_offset_res.to(unit_out).value

    ndim = len(csdm_new.dimensions)

    if dimension_object._complex_fft:
        phase = np.exp(2j * np.pi * coordinates_offset_res * coordinates)
        phase_grid = _get_broadcast_shape(phase, ndim, axis=index)
        for item in csdm_new.dependent_variables:
            phased_signal = item.subtype._components * phase_grid
            ft_shift = np.fft.ifftshift(phased_signal, axes=index)
            signal_ft = np.fft.ifft(ft_shift, axis=index)

            item.subtype._components = signal_ft
        dimension_object._complex_fft = False
    else:  # FFT is false
        phase = np.exp(-2j * np.pi * coordinates_offset * coordinates_res)
        phase_grid = _get_broadcast_shape(phase, ndim, axis=index)
        for item in csdm_new.dependent_variables:
            ft = np.fft.fft(item.subtype._components, axis=index)
            ft_shift = np.fft.fftshift(ft, axes=index)
            signal_ft = ft_shift * phase_grid

            item.subtype._components = signal_ft
        dimension_object._complex_fft = True

    # get the coordinates of the reciprocal dimension.
    dimension_object._swap()
    dimension_object._increment = dimension_object.reciprocal_increment()
    dimension_object._get_coordinates()
    return csdm_new


#
# def _compare_cv_object(cv1, cv2):
#     if cv1.gcv._getparams == cv2.gcv._getparams:
#         return True
#     return False
#
# def _compare_cv_objects(object1, object2):
#     dim1 = len(object1.dimensions)
#     dim2 = len(object2.dimensions)
#     message = (
#         "{0} and {1} do not have the same set of "
#         "controlled variables."
#     ).format(object1.__name__, object2.__name__)
#     if dim1 != dim2:
#         raise Exception(message)
#     for i in range(dim1):
#         if not _compare_cv_object(
#             object1.dimensions[i],
#             object2.dimensions[i]
#         ):
#             raise Exception(message)
#     return True
#
# def _compare_uv(uv1, uv2):
#     # a = {
#     #     'unit': True,
#     #     'quantity_name': True,
#     #     'quantity_type': True
#     # }
#     a = True
#     if uv1.unit.physical_type != uv2.unit.physical_type:
#         a = False
#     if uv1.quantity_name != uv2.quantity_name:
#         raise Exception(
#             (
#                 "Binary operates are not supported for "
#                 "objects with different quantity."
#             )
#         )
#     if uv1.quantity_type != uv2.quantity_type:
#         raise Exception(
#             (
#                 "Binary operates are not supported for "
#                 "objects with different quantity_type."
#             )
#         )
#     return a
#
# def _compare_uv_objects(object1, object2):
#     dim1 = len(object1.dependent_variables)
#     dim2 = len(object2.dependent_variables)
#     for j in range(dim1):
#         for i in range(dim2):
#             a = _compare_uv(
#                 object1.dependent_variables[i],
#                 object2.dependent_variables[j]
#             )



# -*- coding: utf-8 -*-
import numpy as np

import csdmpy as cp
import csdmpy.statistics as stat


def get_gaussian(x, b, c):
    return np.exp(-((x - b) ** 2) / (2 * c ** 2))


def test_01():
    x = np.arange(100) * 2 - 100.0
    gauss = get_gaussian(x, 5.0, 4.0)
    csdm = cp.as_csdm(gauss, unit="T")
    csdm.dimensions[0] = cp.as_dimension(x, unit="µm")

    gauss_integral = 4.0 * np.sqrt(2 * np.pi)

    # integration
    int_csdm = stat.integral(csdm)
    assert np.allclose(int_csdm.value, gauss_integral)
    assert str(int_csdm.unit) == "T um"

    # mean
    mean_csdm = stat.mean(csdm)[0]
    assert np.allclose(mean_csdm.value, 5.0)
    assert str(mean_csdm.unit) == "um"

    # variance
    var_csdm = stat.var(csdm)[0]
    assert np.allclose(var_csdm.value, 16.0)
    assert str(var_csdm.unit) == "um2"

    # standard deviation
    std_csdm = stat.std(csdm)[0]
    assert np.allclose(std_csdm.value, 4.0)
    assert str(std_csdm.unit) == "um"


def test_02():
    x = np.arange(100) - 50.0
    y = np.arange(500) * 0.1 - 25.0
    gauss1 = get_gaussian(x, 5.0, 4.0)[np.newaxis, :]
    gauss2 = get_gaussian(y, 0.50, 0.40)[:, np.newaxis]
    csdm = cp.as_csdm(gauss1 * gauss2, unit="K")
    csdm.dimensions[1] = cp.as_dimension(y, unit="s")
    csdm.dimensions[0] = cp.as_dimension(x, unit="deg")

    gauss_integral = 4.0 * 2 * np.pi * 0.4

    # integration
    int_csdm = stat.integral(csdm)
    assert np.allclose(int_csdm.value, gauss_integral)
    assert str(int_csdm.unit) == "deg K s"

    # mean
    mean_csdm = stat.mean(csdm)
    assert np.allclose(mean_csdm[0].value, 5.0)
    assert np.allclose(mean_csdm[1].value, 0.5)
    assert str(mean_csdm[0].unit) == "deg"
    assert str(mean_csdm[1].unit) == "s"

    # variance
    var_csdm = stat.var(csdm)
    assert np.allclose(var_csdm[0].value, 16.0)
    assert np.allclose(var_csdm[1].value, 0.16)
    assert str(var_csdm[0].unit) == "deg2"
    assert str(var_csdm[1].unit) == "s2"

    # standard deviation
    std_csdm = stat.std(csdm)
    assert np.allclose(std_csdm[0].value, 4.0)
    assert np.allclose(std_csdm[1].value, 0.4)
    assert str(std_csdm[0].unit) == "deg"
    assert str(std_csdm[1].unit) == "s"



# -*- coding: utf-8 -*-
__all__ = ["test01", "test02", "test03", "test04"]

test01 = "https://osu.box.com/shared/static/yis83dhuj1q7v2srw2b1hjnb66d1rw1h.csdf"
test02 = "https://osu.box.com/shared/static/0jsvs9bmme8vf6yn76bctkhfxm4vkcc8.csdf"
test03 = "https://osu.box.com/shared/static/g4tzjpbvh2o8ew9y5imkhzbpyrlkwf8l.csdf"
test04 = "https://osu.box.com/shared/static/1cmop0icxqcphdydwsk9evj7mh0qtzaq.csdf"



# -*- coding: utf-8 -*-
"""The Internal DependentVariable SubType class."""
from __future__ import division
from __future__ import print_function

import numpy as np

from csdmpy.dependent_variables.base_class import BaseDependentVariable
from csdmpy.dependent_variables.decoder import Decoder
from csdmpy.dependent_variables.sparse import SparseSampling
from csdmpy.utils import numpy_dtype_to_numeric_type


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["InternalDataset"]


class InternalDataset(BaseDependentVariable):
    """InternalDataset class."""

    __slots__ = "_sparse_sampling"

    def __init__(self, **kwargs):
        """Initialize."""
        self._sparse_sampling = {}

        components = kwargs["components"]
        if isinstance(components, list) and components != []:
            if isinstance(components[0], np.ndarray):
                components = np.asarray(components)

        if isinstance(components, np.ndarray):
            if kwargs["numeric_type"] is None:
                kwargs["numeric_type"] = numpy_dtype_to_numeric_type(
                    str(components.dtype)
                )
                kwargs["components"] = components
            else:
                kwargs["components"] = components.astype(kwargs["numeric_type"])

        if kwargs["numeric_type"] is None:
            raise KeyError(
                "Missing a required `numeric_type` key from the DependentVariable "
                "object."
            )

        # super base class must be initialized before retrieving
        # the components array.
        super().__init__(**kwargs)

        if not isinstance(components, np.ndarray):
            components = Decoder(
                self._encoding,
                self._quantity_type,
                components,
                self._numeric_type.dtype,
            )
            self._components = components

        size = self._components.size
        p = self.quantity_type.p

        self._components.shape = (p, int(size / p))

        if kwargs["sparse_sampling"] != {}:
            self._sparse_sampling = SparseSampling(**kwargs["sparse_sampling"])

    def __eq__(self, other):
        """Overrides the default implementation"""
        check = [super().__eq__(other), self._sparse_sampling == other._sparse_sampling]
        if False in check:
            return False
        return True

    def to_dict(
        self, filename=None, dataset_index=None, for_display=False, version=None
    ):
        """Alias to the `dict()` method of the class."""
        return self.dict(filename, dataset_index, for_display, version)

    def dict(self, filename=None, dataset_index=None, for_display=False, version=None):
        """Return InternalDataset object as a python dictionary."""
        dictionary = {}

        dictionary["type"] = "internal"
        dictionary.update(
            self._get_dictionary(filename, dataset_index, for_display, version)
        )
        return dictionary



# -*- coding: utf-8 -*-
"""The Base DependentVariable class."""
from __future__ import division
from __future__ import print_function

import base64
import warnings
from copy import deepcopy

import numpy as np

from csdmpy.dependent_variables.download import get_relative_url_path
from csdmpy.units import check_quantity_name
from csdmpy.units import ScalarQuantity
from csdmpy.utils import check_encoding
from csdmpy.utils import NumericType
from csdmpy.utils import numpy_dtype_to_numeric_type
from csdmpy.utils import QuantityType
from csdmpy.utils import validate


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["BaseDependentVariable"]


class BaseDependentVariable:
    r"""Declare a BaseDependentVariable class."""

    __slots__ = (
        "_name",
        "_unit",
        "_quantity_name",
        "_encoding",
        "_numeric_type",
        "_quantity_type",
        "_component_labels",
        "_components",
        "_application",
        "_description",
    )

    def __init__(
        self,
        name="",
        unit="",
        quantity_name=None,
        encoding="none",
        numeric_type="float32",
        quantity_type="scalar",
        components=None,
        component_labels=None,
        description="",
        application={},
        **kwargs,
    ):
        r"""Instantiate a BaseDependentVariable class."""
        self.name = name
        self._unit = ScalarQuantity(f"1 {unit}").quantity.unit
        self._quantity_name = check_quantity_name(quantity_name, self._unit)
        self.encoding = encoding
        self._numeric_type = NumericType(numeric_type)
        self._quantity_type = QuantityType(quantity_type)
        self.set_component_labels(component_labels)
        self.description = description
        self.application = application
        self._components = components

    def __eq__(self, other):
        """Overrides the default implementation"""
        check = [
            self.name == other.name,
            self._unit == other._unit,
            self._quantity_name == other._quantity_name,
            self._encoding == other._encoding,
            self._numeric_type == other._numeric_type,
            self._quantity_type == other._quantity_type,
            self._component_labels == other._component_labels,
            self._description == other._description,
            self._application == other._application,
            np.allclose(self._components, other._components),
        ]
        if False in check:
            return False
        return True

    def set_component_labels(self, component_labels):
        """
        Assign an array of strings, based on the number of components.

        If no label is provided, a default values,
        :math:`['', '', N_k]`, is assigned. If the number of component labels
        does not match the total number of components, a warning is raised and
        the inconsistency is resolved by appropriate truncating or adding the
        required number of strings.
        """
        n = self._quantity_type.p
        if component_labels is None:
            self._component_labels = ["" for i in range(n)]
            return

        validate(component_labels, "component_labels", list)

        component_length = len(component_labels)
        if component_length == n:
            self._component_labels = component_labels
            return

        warning_statement = (
            f"The number of component labels, {component_length}, is not equal"
            f" to the number of components, {n}. The inconsistency is resolved"
            f" by appropriate truncation or addition of the strings."
        )
        warnings.warn(warning_statement)

        if component_length > n:
            self._component_labels = component_labels[:n]
            return

        lables = ["" for i in range(n)]
        for i, item in enumerate(component_labels):
            lables[i] = item
        self._component_labels = lables
        return

    # ----------------------------------------------------------------------- #
    #                                Attributes                               #
    # ----------------------------------------------------------------------- #
    @property
    def name(self):
        """Return name associated with the dependent variable."""
        return deepcopy(self._name)

    @name.setter
    def name(self, value):
        self._name = validate(value, "name", str)

    @property
    def unit(self):
        """Return uint associated with the dependent variable."""
        return deepcopy(self._unit)

    @property
    def quantity_name(self):
        """Return quantity name associated with the physical quantity."""
        return deepcopy(self._quantity_name)

    @quantity_name.setter
    def quantity_name(self, value=""):
        raise NotImplementedError("The `quantity_name` attribute cannot be modified.")

    @property
    def encoding(self):
        r"""Return encoding method used in storing dependent variable."""
        return deepcopy(self._encoding)

    @encoding.setter
    def encoding(self, value):
        self._encoding = validate(value, "encoding", str, method=check_encoding)

    @property
    def numeric_type(self):
        r"""Return numeric type associated with dependent variable."""
        return deepcopy(self._numeric_type)

    @numeric_type.setter
    def numeric_type(self, value):
        self._numeric_type.update(value)

    @property
    def quantity_type(self):
        r"""Return quantity type associated with dependent variable."""
        return deepcopy(self._quantity_type)

    @quantity_type.setter
    def quantity_type(self, value):
        self._quantity_type.update(value)

    @property
    def component_labels(self):
        r"""
        Return an ordered array of labels associated with each component of the
        dependent variable.
        """
        return self._component_labels

    @component_labels.setter
    def component_labels(self, value):
        self.set_component_labels(value)

    @property
    def application(self):
        """Return the application metadata dictionary, if any."""
        return self._application

    @application.setter
    def application(self, value):
        self._application = validate(value, "application", dict)

    @property
    def description(self):
        r"""Return a description of the dependent variable."""
        return deepcopy(self._description)

    @description.setter
    def description(self, value):
        self._description = validate(value, "description", str)

    @property
    def components(self):
        """Return components array."""
        dtype = self._numeric_type.dtype
        if self._components.dtype != dtype:
            self._components = np.asarray(self._components, dtype)
        return self._components

    @components.setter
    def components(self, value):
        value = np.asarray(value)
        if value.shape == self.components.shape:
            self.set_components(value)
            return
        raise ValueError(
            f"The shape of the `{value.__class__.__name__}`, `{value.shape}`, is "
            f"inconsistent with the shape of the components array, "
            f"`{self.components.shape}`."
        )

    # ----------------------------------------------------------------------- #
    #                                  Methods                                #
    # ----------------------------------------------------------------------- #

    def _get_dictionary(
        self, filename=None, dataset_index=None, for_display=False, version=None
    ):
        r"""Return a dictionary object of the base class."""
        obj = {}
        if self._description.strip() != "":
            obj["description"] = str(self._description)

        if self._name.strip() != "":
            obj["name"] = self._name

        if str(self._unit) != "":
            obj["unit"] = ScalarQuantity(1.0 * self._unit).__format__("unit")

        if self._quantity_name not in ["dimensionless", "unknown", None]:
            obj["quantity_name"] = self._quantity_name

        obj["encoding"] = str(self._encoding)
        obj["numeric_type"] = str(self._numeric_type)
        obj["quantity_type"] = str(self._quantity_type)

        # print_label = False
        for label in self._component_labels:
            if label.strip() != "":
                obj["component_labels"] = self._component_labels
                break

        if self._application != {}:
            obj["application"] = self._application

        if for_display:
            obj["components"] = reduced_display(self._components)
            del obj["encoding"]
            return obj

        self.get_proper_encoded_data(obj, filename, dataset_index)

        return obj

    def get_proper_encoded_data(self, obj, filename=None, dataset_index=None):
        c = self.ravel_data()

        if self.encoding == "none":
            obj["components"] = c.tolist()

        if self.encoding == "base64":
            obj["components"] = [base64.b64encode(item).decode("utf-8") for item in c]

        if self.encoding == "raw":
            url_relative_path, absolute_path = get_relative_url_path(
                dataset_index, filename
            )

            c.ravel().tofile(absolute_path)

            obj["type"] = "external"
            obj["components_url"] = url_relative_path
            del obj["encoding"]

        del c
        return obj

    def set_components(self, _components, _numeric_type=None):
        if _numeric_type is None:
            _numeric_type = numpy_dtype_to_numeric_type(str(_components.dtype))
        self._numeric_type.update(_numeric_type)

        self._components = np.asarray(_components, self._numeric_type.dtype)

    def ravel_data(self):
        """Encode data based on the encoding key value."""
        n = self._quantity_type.p
        size = self._components[0].size
        if self._numeric_type.value in ["complex64", "complex128"]:

            if self._numeric_type.value == "complex64":
                c = np.empty((n, size * 2), dtype=np.float32)

            if self._numeric_type.value == "complex128":
                c = np.empty((n, size * 2), dtype=np.float64)

            for i in range(n):
                c[i, 0::2] = self._components.real[i].ravel()
                c[i, 1::2] = self._components.imag[i].ravel()

        else:
            c = np.empty((n, size), dtype=self._numeric_type.dtype)
            for i in range(n):
                c[i] = self._components[i].ravel()

        return c


def reduced_display(_components):
    r"""
        Reduced display for quick view of the data structure.

        The method shows the first and the last two data values.
    """
    _string = []
    for i, item in enumerate(_components):
        temp = item.ravel()
        lst = [str(temp[0]), str(temp[1]), str(temp[-2]), str(temp[-1])]
        _string.append([("{0}, {1}, ..., {2}, {3}").format(*lst)])
    return _string



# -*- coding: utf-8 -*-
"""Test for the csdm object
    1) generate csdm object.
    2) split multiple dependent variables to individual objects.
    3) add, sub, iadd, radd, isub, rsub, mul, imul, for scalar and ScalarQuantity.
    4) rmul, truediv, itruediv, rtruediv, pow, ipow for scalar and ScalarQuantity.
    5) min, max, clip, real, imag, conj, round, angle functions.
"""
import json

import numpy as np
import pytest

import csdmpy as cp


def get_test(type):
    out = np.random.rand(10).astype(type)
    a_test = cp.new()
    a_test.add_dimension(cp.LinearDimension(count=10, increment="1s"))
    a_test.add_dependent_variable(
        {"type": "internal", "quantity_type": "scalar", "unit": "m", "components": out}
    )
    return out, a_test


def get_test_2d(type):
    out = np.random.rand(50).astype(type).reshape(10, 5)
    a_test = cp.new()
    a_test.add_dimension(cp.LinearDimension(count=5, increment="1s"))
    a_test.add_dimension(cp.LinearDimension(count=10, increment="1m"))
    a_test.add_dependent_variable(
        {
            "type": "internal",
            "quantity_type": "scalar",
            "unit": "m",
            "components": out.ravel(),
        }
    )
    return out, a_test


def test_csdm():
    data = cp.new(description="This is a test")

    assert data != "sd"

    assert data.size == 1
    # read_only
    assert data.read_only is False
    data.read_only = True
    assert data.read_only is True
    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.read_only = "True"

    # tags
    assert data.tags == []
    data.tags = ["1", "2", "3"]
    assert data.tags == ["1", "2", "3"]
    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.tags = "23"

    # version
    assert data.version == cp.csdm.CSDM.__latest_CSDM_version__

    # geographic_coordinate
    assert data.geographic_coordinate == {}
    error = "can't set attribute"
    with pytest.raises(AttributeError, match=".*{0}.*".format(error)):
        data.geographic_coordinate = {}

    # description
    assert data.description == "This is a test"
    data.description = "Enough with the tests"
    assert data.description == "Enough with the tests"
    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.description = {}

    # application
    assert data.application == {}
    data.application = {"csdmpy": "Some day"}
    assert data.application == {"csdmpy": "Some day"}
    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        data.application = "Some other day"

    # filename
    assert data.filename == ""

    # data_structure
    structure = {
        "csdm": {
            "version": "1.0",
            "read_only": True,
            "tags": ["1", "2", "3"],
            "description": "Enough with the tests",
            "application": {"csdmpy": "Some day"},
            "dimensions": [],
            "dependent_variables": [],
        }
    }
    assert data.data_structure == str(
        json.dumps(structure, ensure_ascii=False, sort_keys=False, indent=2)
    )

    assert data.dict() == structure

    # equality check
    dm = data.copy()

    assert dm == data
    assert dm.shape == ()

    dm.add_dimension(cp.LinearDimension(count=10, increment="1s"))

    assert dm != data


def test_split():
    a = cp.new()
    a.add_dimension(cp.LinearDimension(count=10, increment="1m"))
    a.add_dependent_variable(
        {"type": "internal", "components": np.arange(10) + 1, "quantity_type": "scalar"}
    )

    b = cp.new()
    b.add_dimension(cp.LinearDimension(count=10, increment="1m"))
    b.add_dependent_variable(
        {"type": "internal", "components": np.arange(10) + 2, "quantity_type": "scalar"}
    )

    c = cp.new()
    c.add_dimension(cp.LinearDimension(count=10, increment="1m"))
    c.add_dependent_variable(
        {"type": "internal", "components": np.arange(10) + 1, "quantity_type": "scalar"}
    )
    c.add_dependent_variable(
        {"type": "internal", "components": np.arange(10) + 2, "quantity_type": "scalar"}
    )

    a_, b_ = c.split()

    assert a_ == a
    assert b_ == b


a_test = cp.new()
a_test.add_dimension(cp.LinearDimension(count=10, increment="1s"))
a_test.add_dependent_variable(
    {
        "type": "internal",
        "quantity_type": "scalar",
        "unit": "m",
        "components": np.arange(10),
    }
)

a1_test = cp.new()
a1_test.add_dimension(cp.LinearDimension(count=10, increment="1m"))
a1_test.add_dependent_variable(
    {
        "type": "internal",
        "quantity_type": "scalar",
        "unit": "m",
        "components": np.arange(10),
    }
)

b_test = cp.new()
b_test.add_dimension(cp.LinearDimension(count=10, increment="1s"))
b_test.add_dependent_variable(
    {
        "type": "internal",
        "quantity_type": "scalar",
        "unit": "km",
        "components": np.arange(10, dtype=float),
    }
)

b1_test = cp.new()
b1_test.add_dimension(cp.LinearDimension(count=10, increment="1s"))
b1_test.add_dependent_variable(
    {
        "type": "internal",
        "quantity_type": "scalar",
        "unit": "km",
        "components": np.arange(10),
    }
)
b1_test.add_dependent_variable(
    {
        "type": "internal",
        "quantity_type": "vector_2",
        "unit": "km",
        "components": np.arange(20),
    }
)


def test_add_sub():
    # add
    c = a_test + b_test
    out = np.arange(10) + 1000 * np.arange(10)
    assert np.allclose(c.dependent_variables[0].components, [out])

    # sub
    c = a_test - b_test
    out = np.arange(10) - 1000 * np.arange(10)
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = b_test - a_test
    out = np.arange(10) - 1 / 1000 * np.arange(10)
    assert np.allclose(c.dependent_variables[0].components, [out])

    # add scalar
    c = a_test + cp.ScalarQuantity("2m")
    out = np.arange(10) + 2
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = cp.ScalarQuantity("2m") + a_test
    assert np.allclose(c.dependent_variables[0].components, [out])

    a_t = cp.as_csdm(np.arange(10))
    c = a_t + 5.12
    out = np.arange(10) + 5.12
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = 5.12 + a_t
    assert np.allclose(c.dependent_variables[0].components, [out])

    # add complex
    c = a_t + 5.12 - 4j
    out = np.arange(10) + 5.12 - 4j
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = 5.12 - 4j + a_t
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = a_test / cp.ScalarQuantity("m") + 3 + 4j
    out = np.arange(10) + 3 + 4j
    assert np.allclose(c.dependent_variables[0].components, [out])

    # subtract scalar
    c = a_test - cp.ScalarQuantity("2m")
    out = np.arange(10) - 2
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = cp.ScalarQuantity("2m") - a_test
    assert np.allclose(c.dependent_variables[0].components, [-out])

    error = r"Cannot operate on CSDM objects with different dimensions."
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        c = a1_test + b_test

    error = r"Cannot operate on CSDM objects with differnet lengths of dependent"
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        c = a_test + b1_test


def test_iadd_isub():
    c = a_test.astype("float32")
    c += cp.ScalarQuantity("5.0cm")
    out = np.arange(10) + 0.05
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = a_test.astype("float32")
    c -= cp.ScalarQuantity("5.0cm")
    out = np.arange(10) - 0.05
    assert np.allclose(c.dependent_variables[0].components, [out])

    c = a_test.astype("float32") / cp.ScalarQuantity("cm")
    c -= 0.05
    out = np.arange(10) - 0.05
    assert np.allclose(c.dependent_variables[0].components, [out])

    out, a_test_ = get_test(float)  # in units of m
    a_test_ += b_test
    out += 1000 * np.arange(10)
    assert np.allclose(a_test_.dependent_variables[0].components, [out])

    out, a_test_ = get_test(float)  # in units of m
    c = a_test_ / cp.ScalarQuantity("m")
    c += 2.0
    out += 2.0
    assert np.allclose(c.dependent_variables[0].components, [out])

    out, a_test_ = get_test(float)  # in units of m
    a_test_ -= b_test
    out -= 1000 * np.arange(10)
    assert np.allclose(a_test_.dependent_variables[0].components, [out])


def test_mul_truediv_pow():
    # mul
    c = a_test * 2
    out = a_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out * 2)
    assert id(a_test) != id(c)

    c = 2 * a_test
    assert np.allclose(c.dependent_variables[0].components, out * 2)

    c = b_test * 1.34
    out = b_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out * 1.34)
    assert id(b_test) != id(c)

    c = 1.34 * a_test
    assert np.allclose(c.dependent_variables[0].components, out * 1.34)

    c = b1_test * 1.423j
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out * 1.423j)
    assert id(b1_test) != id(c)

    c = 1.423j * b1_test
    assert np.allclose(c.dependent_variables[0].components, out * 1.423j)

    lst = np.random.rand(5).astype(np.complex128)
    c = a1_test * lst[2]
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out * lst[2])
    assert id(a1_test) != id(c)

    # c = lst[2] * a1_test
    # assert np.allclose(c.dependent_variables[0].components, out * lst[2])

    c = b1_test * cp.string_to_quantity("1.324 s")
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out * 1.324)
    assert str(c.dependent_variables[0].unit) == "km s"
    assert id(b1_test) != id(c)

    c = b1_test * cp.ScalarQuantity("1.324 s")
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out * 1.324)
    assert str(c.dependent_variables[0].unit) == "km s"
    assert id(b1_test) != id(c)

    c = cp.ScalarQuantity("1.324 s") * b1_test
    assert np.allclose(c.dependent_variables[0].components, out * 1.324)

    # div
    c = a_test / 2
    out = a_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out / 2)
    assert id(a_test) != id(c)

    c = 2.0 / a_test.astype(float)
    assert np.allclose(c.dependent_variables[0].components, 2 / out)

    c = b_test / 1.34
    out = b_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out / 1.34)
    assert id(b_test) != id(c)

    c = b1_test / 1.423j
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out / 1.423j)
    assert id(b1_test) != id(c)

    c = 1.423j / (b1_test + 1).astype(float)
    assert np.allclose(c.dependent_variables[0].components, 1.423j / (out + 1))

    lst = np.random.rand(5).astype(np.complex128)
    c = a1_test / lst[2]
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out / lst[2])
    assert id(a1_test) != id(c)

    c = b1_test / cp.string_to_quantity("1.324 s")
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out / 1.324)
    assert str(c.dependent_variables[0].unit) == "km / s"
    assert id(b1_test) != id(c)

    # c = cp.string_to_quantity("1.324 s") / (b1_test+1)
    # assert np.allclose(c.dependent_variables[0].components, 1.324 / (out+1))

    c = b1_test / cp.ScalarQuantity("1.324 s")
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, out / 1.324)
    assert str(c.dependent_variables[0].unit) == "km / s"
    assert id(b1_test) != id(c)

    c = cp.ScalarQuantity("1.324 s") / (b_test + 1)
    assert np.allclose(c.dependent_variables[0].components, 1.324 / (out + 1))

    error = r"unsupported operand type\(s\) \*: 'CSDM' and 'str'."
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        c = a_test * "3"

    error = r"Only scalar multiplication or division is allowed."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        c = a_test * np.asarray([1, 2])

    error = r"unsupported operand type\(s\) /: 'CSDM' and 'str'."
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        c = a_test / "3"

    error = r"Only scalar multiplication or division is allowed."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        c = a_test / np.asarray([1, 2])

    # pow
    c = b1_test ** 2
    out = b1_test.dependent_variables[0].components
    assert np.allclose(c.dependent_variables[0].components, np.power(out, 2))
    assert str(c.dependent_variables[0].unit) == "km2"

    c **= 2
    assert np.allclose(c.dependent_variables[0].components, np.power(out, 4))
    assert str(c.dependent_variables[0].unit) == "km4"

    c **= 1 / 4
    assert np.allclose(c.dependent_variables[0].components, out)
    assert str(c.dependent_variables[0].unit) == "km"


def test_imul_itruediv():
    # imul
    out, a_test = get_test(int)
    b = a_test
    a_test *= 2
    out *= 2
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert id(a_test) == id(b)

    out, a_test = get_test(float)
    b = a_test
    a_test *= 2.312
    out *= 2.312
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert id(a_test) == id(b)

    out, a_test = get_test(complex)
    b = a_test
    a_test *= 2.12 + 1.1j
    out *= 2.12 + 1.1j
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert id(a_test) == id(b)

    out, a_test = get_test(float)
    b = a_test
    a_test *= cp.string_to_quantity("3.12 s/km")
    out *= 3.12
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert str(a_test.dependent_variables[0].unit) == "m s / km"
    assert id(a_test) == id(b)

    out, a_test = get_test(float)
    b = a_test
    a_test *= cp.ScalarQuantity("3.12 s/km")
    out *= 3.12
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert str(a_test.dependent_variables[0].unit) == "m s / km"
    assert id(a_test) == id(b)

    # itruediv

    out, a_test = get_test(float)
    b = a_test
    a_test /= 2
    out /= 2
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert id(a_test) == id(b)

    out, a_test = get_test(float)
    b = a_test
    a_test /= 2.312
    out /= 2.312
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert id(a_test) == id(b)

    out, a_test = get_test(complex)
    b = a_test
    a_test /= 2.12 + 1.1j
    out /= 2.12 + 1.1j
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert id(a_test) == id(b)

    out, a_test = get_test(float)
    b = a_test
    a_test /= cp.string_to_quantity("3.12 s/km")
    out /= 3.12
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert str(a_test.dependent_variables[0].unit) == "km m / s"
    assert id(a_test) == id(b)

    out, a_test = get_test(float)
    b = a_test
    a_test /= cp.ScalarQuantity("3.12 s/km")
    out /= 3.12
    assert np.allclose(a_test.dependent_variables[0].components, [out])
    assert str(a_test.dependent_variables[0].unit) == "km m / s"
    assert id(a_test) == id(b)


def test_in_build():
    assert a_test == +a_test

    out = a_test.dependent_variables[0].components
    assert np.allclose(abs(a_test).dependent_variables[0].components[0], abs(out))


def test_max_min_clip():
    # max
    d0 = cp.LinearDimension(count=5, increment="1s")
    d1 = cp.LinearDimension(count=10, increment="1m")
    out, a_test = get_test_2d(float)
    assert out.max() == a_test.max().value
    assert str(a_test.max().unit) == "m"

    b = a_test.max(axis=1)
    assert np.allclose(out.max(0), b.dependent_variables[0].components[0])
    assert b.dimensions[0] == d0

    b = a_test.max(axis=0)
    assert np.allclose(out.max(1), b.dependent_variables[0].components[0])
    assert b.dimensions[0] == d1

    # min
    out, a_test = get_test_2d(float)
    assert out.min() == a_test.min().value
    assert str(a_test.min().unit) == "m"

    b = a_test.min(axis=1)
    assert np.allclose(out.min(0), b.dependent_variables[0].components[0])
    assert b.dimensions[0] == d0

    b = a_test.min(axis=0)
    assert np.allclose(out.min(1), b.dependent_variables[0].components[0])
    assert b.dimensions[0] == d1

    # clip
    b = a_test.clip(min=0)
    assert np.allclose(out.clip(min=0), b.dependent_variables[0].components[0])

    b = a_test.clip(max=0.5)
    assert np.allclose(out.clip(max=0.5), b.dependent_variables[0].components[0])


def test_real_imag_conj_angle():
    out, a_test = get_test_2d(complex)
    b = a_test.real
    assert b.dependent_variables[0].numeric_type == "float64"
    assert np.allclose(out.real, b.dependent_variables[0].components[0])

    b = np.real(a_test)
    assert b.dependent_variables[0].numeric_type == "float64"
    assert np.allclose(out.real, b.dependent_variables[0].components[0])

    b = a_test.imag
    assert b.dependent_variables[0].numeric_type == "float64"
    assert np.allclose(out.imag, b.dependent_variables[0].components[0])

    b = a_test.conj()
    assert b.dependent_variables[0].numeric_type == "complex128"
    assert np.allclose(out.conj(), b.dependent_variables[0].components[0])

    b = np.angle(a_test)
    assert b.dependent_variables[0].numeric_type == "float64"
    assert np.allclose(np.angle(out), b.dependent_variables[0].components[0])


def test_round_around():
    out, a_test = get_test_2d(float)
    b = a_test.round(1)
    assert np.allclose(out.round(1), b.dependent_variables[0].components[0])

    b = np.around(a_test, 1)
    assert np.allclose(np.around(out, 1), b.dependent_variables[0].components[0])


def test_not_implemented_error():
    fn = [np.argmax, np.argmin, np.ptp, np.trace, np.cumsum, np.cumprod]
    for fn_ in fn:
        with pytest.raises(NotImplementedError, match=""):
            fn_(a_test)

    with pytest.raises(NotImplementedError, match=""):
        a_test.argmax()

    with pytest.raises(NotImplementedError, match=""):
        a_test.argmin()

    with pytest.raises(NotImplementedError, match=""):
        a_test.ptp()

    with pytest.raises(NotImplementedError, match=""):
        a_test.trace()

    with pytest.raises(NotImplementedError, match=""):
        a_test.cumsum()

    with pytest.raises(NotImplementedError, match=""):
        a_test.cumprod()



# -*- coding: utf-8 -*-
from collections.abc import MutableSequence

import numpy as np

from csdmpy.dependent_variables import DependentVariable
from csdmpy.dimensions import Dimension
from csdmpy.dimensions import LabeledDimension
from csdmpy.dimensions import LinearDimension
from csdmpy.dimensions import MonotonicDimension

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

__all__ = ["DimensionList", "DependentVariableList"]

__dimensions_list__ = (Dimension, LinearDimension, MonotonicDimension, LabeledDimension)


class AbstractList(MutableSequence):
    def __init__(self, data=[]):
        super().__init__()
        self._list = list(data)

    def __repr__(self):
        """String representation"""
        return self._list.__repr__()

    def __str__(self):
        """String representation"""
        string = ",\n".join([item.__repr__() for item in self._list])
        return f"[{string}]"

    def __len__(self):
        """List length"""
        return len(self._list)

    def __getitem__(self, index):
        """Get a list item"""
        return self._list[index]

    def __delitem__(self, index):
        raise LookupError("Deleting items is not allowed.")
        # del self._list[index]

    def insert(self, index, item):
        """Insert a list item"""
        self._list.insert(index, item)

    def append(self, item):
        """Append a list item"""
        self.insert(len(self._list), item)

    def __eq__(self, other):
        """Check equality of DependentVariableList."""
        if not isinstance(other, self.__class__):
            return False
        if len(self._list) != len(other._list):
            return False

        check = []
        for i in range(len(self._list)):
            check.append(self._list[i] == other._list[i])

        if np.all(check):
            return True
        return False


class DimensionList(AbstractList):
    def __init__(self, data=[]):
        super().__init__(data)

    def __setitem__(self, index, item):
        if not isinstance(item, __dimensions_list__):
            raise ValueError(
                f"Expecting a Dimension object, found {item.__class__.__name__}"
            )
        if self._list[index].count != item.count:
            raise ValueError(
                f"Cannot substitute a dimension with {self._list[index].count} points "
                f"with a dimension with {item.count} points."
            )
        self._list[index] = item


class DependentVariableList(AbstractList):
    def __init__(self, data=[]):
        super().__init__(data)

    def __setitem__(self, index, item):
        if not isinstance(item, DependentVariable):
            raise ValueError(
                f"Expecting a DependentVariable object, found {item.__class__.name__}"
            )
        self._list[index] = item



# -*- coding: utf-8 -*-
"""Helper functions."""
from copy import deepcopy
from warnings import warn

import matplotlib.projections as proj
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.image import NonUniformImage

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

scalar = ["scalar", "vector_1", "pixel_1", "matrix_1_1", "symmetric_matrix_1"]


def _get_label_from_dv(dv, i):
    """Return label along with the unit of the dependent variable

    Args:
        dv: DependentVariable object.
        i: integer counter.
    """
    name, unit = dv.name, dv.unit
    name = name if name != "" else str(i)
    label = f"{name} / ({unit})" if unit != "" else name
    return label


class CSDMAxes(plt.Axes):
    """A custom CSDM data plot axes."""

    name = "csdm"

    def plot(self, csdm, *args, **kwargs):
        """Produce a figure using the `plot` method from the matplotlib library.

        Apply to all 1D datasets with single-component dependent-variables. For
        multiple dependent variables, the data from individual dependent-variables is
        plotted on the same figure.

        Args:
            csdm: A CSDM object of a one-dimensional dataset.
            kwargs: Additional keyword arguments for the matplotlib plot() method.

        Example
        -------

        >>> ax = plt.subplot(projection='csdm') # doctest: +SKIP
        >>> ax.plot(csdm_object) # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
        """
        if csdm.__class__.__name__ != "CSDM":
            return super().plot(csdm, *args, **kwargs)

        return self._call_1D(csdm, "plot", *args, **kwargs)

    def scatter(self, csdm, *args, **kwargs):
        """Produce a figure using the `scatter` method from the matplotlib library.

        Apply to all 1D datasets with single-component dependent-variables. For
        multiple dependent variables, the data from individual dependent-variables is
        plotted on the same figure.

        Args:
            csdm: A CSDM object of a one-dimensional dataset.
            kwargs: Additional keyword arguments for the matplotlib plot() method.

        Example
        -------

        >>> ax = plt.subplot(projection='csdm') # doctest: +SKIP
        >>> ax.scatter(csdm_object) # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
        """
        if csdm.__class__.__name__ != "CSDM":
            return super().scatter(csdm, *args, **kwargs)

        return self._call_1D(csdm, "scatter", *args, **kwargs)

    def imshow(self, csdm, origin="lower", *args, **kwargs):
        """Produce a figure using the `imshow` method from the matplotlib library.

        Apply to all 2D datasets with either single-component (scalar),
        three-components (pixel_3), or four-components (pixel_4) dependent-variables.
        For single-component (scalar) dependent-variable, a colormap image is produced.
        For three-components (pixel_3) dependent-variable, an RGB image is produced.
        For four-components (pixel_4) dependent-variable, an RGBA image is produced.

        For multiple dependent variables, the data from individual dependent-variables
        is plotted on the same figure.

        Args:
            csdm: A CSDM object of a two-dimensional dataset with scalar, pixel_3, or
                pixel_4 quantity_type dependent variable.
            origin: The matplotlib `origin` argument. In matplotlib, the default is
                'upper'. In csdmpy, however, the default to 'lower'.
            kwargs: Additional keyword arguments for the matplotlib imshow() method.

        Example
        -------

        >>> ax = plt.subplot(projection='csdm') # doctest: +SKIP
        >>> ax.imshow(csdm_object) # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP

        """
        if csdm.__class__.__name__ != "CSDM":
            return super().imshow(csdm, *args, **kwargs)

        x = csdm.dimensions

        if x[0].type == "linear" and x[1].type == "linear":
            return self._call_uniform_2D_image(csdm, origin=origin, *args, **kwargs)

    def contour(self, csdm, *args, **kwargs):
        """Produce a figure using the `contour` method from the matplotlib library.

        Apply to all 2D datasets with a single-component (scalar) dependent-variables.
        For multiple dependent variables, the data from individual dependent-variables
        is plotted on the same figure.

        Args:
            csdm: A CSDM object of a two-dimensional dataset with scalar dependent
                variable.
            kwargs: Additional keyword arguments for the matplotlib contour() method.

        Example
        -------

        >>> ax = plt.subplot(projection='csdm') # doctest: +SKIP
        >>> ax.contour(csdm_object) # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP

        """
        if csdm.__class__.__name__ != "CSDM":
            return super().contour(csdm, *args, **kwargs)

        x = csdm.dimensions

        if x[0].type == "linear" and x[1].type == "linear":
            return self._call_uniform_2D_contour(csdm, "contour", *args, **kwargs)

    def contourf(self, csdm, *args, **kwargs):
        """Produce a figure using the `contourf` method from the matplotlib library.

        Apply to all 2D datasets with a single-component (scalar) dependent-variables.
        For multiple dependent variables, the data from individual dependent-variables
        is plotted on the same figure.

        Args:
            csdm: A CSDM object of a two-dimensional dataset with scalar dependent
                variable.
            kwargs: Additional keyword arguments for the matplotlib contourf() method.

        Example
        -------

        >>> ax = plt.subplot(projection='csdm') # doctest: +SKIP
        >>> ax.contourf(csdm_object) # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
        """
        if csdm.__class__.__name__ != "CSDM":
            return super().contour(csdm, *args, **kwargs)

        x = csdm.dimensions

        if x[0].type == "linear" and x[1].type == "linear":
            return self._call_uniform_2D_contour(csdm, "contourf", *args, **kwargs)

    def _call_1D(self, csdm, fn, *args, **kwargs):
        _check_1D_dataset(csdm)
        x = csdm.dimensions
        z = csdm.split()
        one = True if len(z) == 1 else False
        legend = False
        for i, item in enumerate(z):
            x_, y_ = item.to_list()
            # dv will always be at index 0 because we called the object.split() before.
            dv = item.dependent_variables[0]

            kwargs_ = deepcopy(kwargs)
            # add a default label if not provided by the user.
            if "label" not in kwargs_.keys():
                kwargs_["label"] = dv.name if one else _get_label_from_dv(dv, i)
                if kwargs_["label"] != "":
                    legend = True

            if fn == "plot":
                r_plt = super().plot(x_, y_, *args, **kwargs_)
            if fn == "scatter":
                r_plt = super().scatter(x_, y_, *args, **kwargs_)

        self.set_xlim(x[0].coordinates.value.min(), x[0].coordinates.value.max())
        self.set_xlabel(x[0].axis_label)

        ylabel = dv.axis_label[0] if one else "dimensionless"
        self.set_ylabel(ylabel)
        self.grid(color="gray", linestyle="--", linewidth=0.5)

        if legend:
            self.legend()

        return r_plt

    def _call_uniform_2D_contour(self, csdm, fn, *args, **kwargs):
        _check_2D_scalar_dataset(csdm)
        kw_keys = kwargs.keys()

        # set extent
        x = csdm.dimensions
        x0, x1 = x[0].coordinates.value, x[1].coordinates.value

        # add cmap for multiple dependent variables.
        cmaps_bool = False
        if "cmaps" in kw_keys:
            cmaps_bool = True
            cmaps = kwargs.pop("cmaps")

        one = True if len(csdm.dependent_variables) == 1 else False

        for i, dv in enumerate(csdm.dependent_variables):
            y = dv.components
            if dv.quantity_type == "scalar":
                if cmaps_bool:
                    kwargs["cmap"] = cmaps[i]

                if fn == "contour":
                    r_plt = super().contour(x0, x1, y[0], *args, **kwargs)
                if fn == "contourf":
                    r_plt = super().contourf(x0, x1, y[0], *args, **kwargs)

                # label = dv.axis_label[0] if one else f"{dv.name} - {dv.axis_label[0]}"
                # cbar = plt.gcf().colorbar(r_plt, ax=self)
                # cbar.ax.minorticks_off()
                # cbar.set_label(label)

        self.set_xlim(x0.min(), x0.max())
        self.set_ylim(x1.min(), x1.max())
        self.set_xlabel(x[0].axis_label)
        self.set_ylabel(x[1].axis_label)
        if one:
            self.set_title(dv.name)
        self.grid(color="gray", linestyle="--", linewidth=0.5)

        return r_plt

    def _call_uniform_2D_image(self, csdm, *args, **kwargs):
        _check_2D_scalar_and_pixel_dataset(csdm)

        kw_keys = kwargs.keys()

        # set extent
        x = csdm.dimensions
        x0, x1 = x[0].coordinates.value, x[1].coordinates.value
        extent = [x0[0], x0[-1], x1[0], x1[-1]]
        if kwargs["origin"] == "upper":
            extent = [x0[0], x0[-1], x1[-1], x1[0]]
        if "extent" not in kw_keys:
            kwargs["extent"] = extent

        # add cmap for multiple dependent variables.
        cmaps_bool = False
        if "cmaps" in kw_keys:
            cmaps_bool = True
            cmaps = kwargs.pop("cmaps")

        one = True if len(csdm.dependent_variables) == 1 else False

        for i, dv in enumerate(csdm.dependent_variables):
            y = dv.components
            if dv.quantity_type == "scalar":
                if cmaps_bool:
                    kwargs["cmap"] = cmaps[i]

                r_plt = super().imshow(y[0], *args, **kwargs)

                # label = dv.axis_label[0] if one else f"{dv.name} - {dv.axis_label[0]}"
                # cbar = plt.gcf().colorbar(r_plt, ax=self)
                # cbar.ax.minorticks_off()
                # cbar.set_label(label)

            if dv.quantity_type == "pixel_3":
                r_plt = super().imshow(np.moveaxis(y.copy(), 0, -1), *args, **kwargs)

            if dv.quantity_type == "pixel_4":
                r_plt = super().imshow(np.moveaxis(y.copy(), 0, -1), *args, **kwargs)

        self.set_xlabel(x[0].axis_label)
        self.set_ylabel(x[1].axis_label)
        if one:
            self.set_title(dv.name)
        self.grid(color="gray", linestyle="--", linewidth=0.5)

        return r_plt


proj.register_projection(CSDMAxes)


def _check_1D_dataset(csdm):
    x, y = csdm.dimensions, csdm.dependent_variables

    message = (
        "The function requires a 1D dataset with single-component dependent "
        "variables. For multiple dependent-variables, the data from all the "
        "dependent variables are ploted on the same figure."
    )
    if len(x) != 1:
        raise Exception(message)
    for y_ in y:
        if len(y_.components) != 1:
            raise Exception(message)


def _check_2D_scalar_and_pixel_dataset(csdm):
    x, y = csdm.dimensions, csdm.dependent_variables

    message = (
        "The function requires a 2D dataset with a single-component (scalar), "
        "three components (pixel_3), or four components (pixel_4) dependent "
        "variables. The pixel_3 produces an RGB image while pixel_4, a RGBA image."
    )
    if len(x) != 2:
        raise Exception(message)
    for y_ in y:
        if len(y_.components) not in [1, 3, 4]:
            raise Exception(message)


def _check_2D_scalar_dataset(csdm):
    x, y = csdm.dimensions, csdm.dependent_variables

    message = (
        "The function requires a 2D dataset with a single-component (scalar), "
        "dependent variables."
    )
    if len(x) != 2:
        raise Exception(message)
    for y_ in y:
        if len(y_.components) != 1:
            raise Exception(message)


# --------- cp plot functions ---------- #


def _preview(data, reverse_axis=None, range_=None, **kwargs):
    """Quick display of the data."""
    if reverse_axis is not None:
        kwargs["reverse_axis"] = reverse_axis

    if range_ is None:
        range_ = [[None, None], [None, None]]

    x = data.dimensions
    y = data.dependent_variables
    y_len = len(y)
    y_grid = int(y_len / 2) + 1

    if len(x) == 0:
        raise NotImplementedError(
            "Preview of zero dimensional datasets is not implemented."
        )

    if len(x) > 2:
        raise NotImplementedError(
            "Preview of three or higher dimensional datasets " "is not implemented."
        )

    if np.any([x[i].type == "labeled" for i in range(len(x))]):
        raise NotImplementedError("Preview of labeled dimensions is not implemented.")

    fig = plt.gcf()
    if y_len <= 2:
        ax = fig.subplots(y_grid)
        ax = [[ax]] if y_len == 1 else [ax]
    else:
        ax = fig.subplots(y_grid, 2)

    if len(x) == 1:
        one_d_plots(ax, x, y, range_, **kwargs)

    if len(x) == 2:
        two_d_plots(ax, x, y, range_, **kwargs)

    return fig


def one_d_plots(ax, x, y, range_, **kwargs):
    """A collection of possible 1D plots."""
    for i, y_item in enumerate(y):
        i0 = int(i / 2)
        j0 = int(i % 2)
        ax_ = ax[i0][j0]

        if y_item.quantity_type in scalar:
            oneD_scalar(x, y_item, ax_, range_, **kwargs)
        if "vector" in y_item.quantity_type:
            vector_plot(x, y_item, ax_, range_, **kwargs)
        # if "audio" in y_item.quantity_type:
        #     audio(x, y, i, fig, ax, **kwargs)


def two_d_plots(ax, x, y, range_, **kwargs):
    """A collection of possible 2D plots."""
    for i, y_item in enumerate(y):
        i0 = int(i / 2)
        j0 = int(i % 2)
        ax_ = ax[i0][j0]

        if y_item.quantity_type == "pixel_3":
            warn("This method interprets the `pixel_3` dataset as an RGB image.")
            RGB_image(x, y_item, ax_, range_, **kwargs)

        if y_item.quantity_type in scalar:
            twoD_scalar(x, y_item, ax_, range_, **kwargs)

        if "vector" in y_item.quantity_type:
            vector_plot(x, y_item, ax_, range_, **kwargs)


def oneD_scalar(x, y, ax, range_, **kwargs):
    reverse = [False]
    if "reverse_axis" in kwargs.keys():
        reverse = kwargs["reverse_axis"]
        kwargs.pop("reverse_axis")

    components = y.components.shape[0]
    for k in range(components):
        ax.plot(x[0].coordinates, y.components[k], **kwargs)
        ax.set_xlim(x[0].coordinates.value.min(), x[0].coordinates.value.max())
        ax.set_xlabel(f"{x[0].axis_label} - 0")
        ax.set_ylabel(y.axis_label[0])
        ax.set_title("{0}".format(y.name))
        ax.grid(color="gray", linestyle="--", linewidth=0.5)

    ax.set_xlim(range_[0])
    ax.set_ylim(range_[1])

    if reverse[0]:
        ax.invert_xaxis()


def twoD_scalar(x, y, ax, range_, **kwargs):
    reverse = [False, False]
    if "reverse_axis" in kwargs.keys():
        reverse = kwargs["reverse_axis"]
        kwargs.pop("reverse_axis")

    x0 = x[0].coordinates.value
    x1 = x[1].coordinates.value
    y00 = y.components[0]
    extent = [x0[0], x0[-1], x1[0], x1[-1]]
    if "extent" not in kwargs.keys():
        kwargs["extent"] = extent

    if x[0].type == "linear" and x[1].type == "linear":
        if "origin" not in kwargs.keys():
            kwargs["origin"] = "lower"
        if "aspect" not in kwargs.keys():
            kwargs["aspect"] = "auto"

        cs = ax.imshow(y00, **kwargs)
    else:
        if "interpolation" not in kwargs.keys():
            kwargs["interpolation"] = "nearest"

        cs = NonUniformImage(ax, **kwargs)
        cs.set_data(x0, x1, y00)
        ax.images.append(cs)

    cbar = ax.figure.colorbar(cs, ax=ax)
    cbar.ax.minorticks_off()
    cbar.set_label(y.axis_label[0])
    ax.set_xlim([extent[0], extent[1]])
    ax.set_ylim([extent[2], extent[3]])
    ax.set_xlabel(f"{x[0].axis_label} - 0")
    ax.set_ylabel(f"{x[1].axis_label} - 1")
    ax.set_title("{0}".format(y.name))
    ax.grid(color="gray", linestyle="--", linewidth=0.5)

    ax.set_xlim(range_[0])
    ax.set_ylim(range_[1])

    if reverse[0]:
        ax.invert_xaxis()
    if reverse[1]:
        ax.invert_yaxis()


def vector_plot(x, y, ax, range_, **kwargs):
    reverse = [False, False]
    if "reverse_axis" in kwargs.keys():
        reverse = kwargs["reverse_axis"]
        kwargs.pop("reverse_axis")

    x0 = x[0].coordinates.value
    if len(x) == 2:
        x1 = x[1].coordinates.value
    else:
        x1 = np.zeros(1)

    x0, x1 = np.meshgrid(x0, x1)
    u1 = y.components[0]
    v1 = y.components[1]

    if "pivot" not in kwargs.keys():
        kwargs["pivot"] = "middle"
    ax.quiver(x0, x1, u1, v1, **kwargs)
    ax.set_xlabel(f"{x[0].axis_label} - 0")
    ax.set_xlim(x[0].coordinates.value.min(), x[0].coordinates.value.max())
    if len(x) == 2:
        ax.set_ylim(x[1].coordinates.value.min(), x[1].coordinates.value.max())
        ax.set_ylabel(f"{x[1].axis_label} - 1")
        if reverse[1]:
            ax.invert_yaxis()
    else:
        ax.set_ylim([-y.components.max(), y.components.max()])
    ax.set_title("{0}".format(y.name))
    ax.grid(color="gray", linestyle="--", linewidth=0.5)

    ax.set_xlim(range_[0])
    ax.set_ylim(range_[1])

    if reverse[0]:
        ax.invert_xaxis()


def RGB_image(x, y, ax, range_, **kwargs):
    reverse = [False, False]
    if "reverse_axis" in kwargs.keys():
        reverse = kwargs["reverse_axis"]
        kwargs.pop("reverse_axis")

    y0 = y.components
    ax.imshow(np.moveaxis(y0 / y0.max(), 0, -1), **kwargs)
    ax.set_title("{0}".format(y.name))

    ax.set_xlim(range_[0])
    ax.set_ylim(range_[1])

    if reverse[0]:
        ax.invert_xaxis()
    if reverse[1]:
        ax.invert_yaxis()


# def audio(x, y, i0, fig, ax):
#     try:
#         SOUND = 1
#         import sounddevice as sd
#     except ImportError:
#         SOUND = 0
#         string = (
#             "Module 'sounddevice' is not installed. All audio data files will "
#             "not be played. To enable audio files, install 'sounddevice' using"
#             " 'pip install sounddevice'."
#         )
#         warn(string)

#     plot1D(x, y, i0, ax)
#     if SOUND == 1:
#         data_max = y[i0].components.max()
#         sd.play(0.9 * y[i0].components.T / data_max, 1 / x[0].increment.to("s").value)



# -*- coding: utf-8 -*-
"""CSDM main."""
import sys

import matplotlib.pyplot as plt

import csdmpy as cp
from csdmpy.helper_functions import _preview

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


if __name__ == "__main__":
    for i in range(len(sys.argv) - 1):
        data = cp.load(sys.argv[i + 1])
        _preview(data)
        plt.show()



# -*- coding: utf-8 -*-
"""The MonotonicDimension sub type class."""
from __future__ import division
from __future__ import print_function

import json
from copy import deepcopy

import numpy as np
from astropy.units import Quantity

from csdmpy.dimensions.base import _copy_core_metadata
from csdmpy.dimensions.base import check_count
from csdmpy.dimensions.quantitative import BaseQuantitativeDimension
from csdmpy.dimensions.quantitative import ReciprocalDimension
from csdmpy.units import frequency_ratio
from csdmpy.units import scalar_quantity_format
from csdmpy.units import ScalarQuantity
from csdmpy.utils import _axis_label
from csdmpy.utils import attribute_error
from csdmpy.utils import check_scalar_object


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["MonotonicDimension"]

# =========================================================================== #
#                     MonotonicDimension Class                       #
# =========================================================================== #


class MonotonicDimension(BaseQuantitativeDimension):
    r"""
    A monotonic grid dimension.

    Generates an object representing a physical dimension whose coordinates are
    monotonically sampled along a grid dimension. See :ref:`monotonicDimension_uml`
    for details.
    """

    __slots__ = ("reciprocal", "_count", "_values", "_coordinates")

    _type = "monotonic"

    def __init__(self, coordinates, **kwargs):
        """Instantiate a MonotonicDimension class."""
        if isinstance(coordinates, Quantity):
            _unit = coordinates.unit
        elif isinstance(coordinates, np.ndarray):
            _unit = Quantity(1, "").unit
        else:
            _unit = ScalarQuantity(coordinates[0]).quantity.unit
        if "reciprocal" not in kwargs.keys():
            kwargs["reciprocal"] = {
                "increment": None,
                "coordinates_offset": None,
                "origin_offset": None,
                "period": None,
                "quantity_name": None,
                "label": "",
                "description": "",
                "application": {},
            }
        super().__init__(unit=_unit, **kwargs)

        _reciprocal_unit = self._unit ** -1
        self.reciprocal = ReciprocalDimension(
            unit=_reciprocal_unit, **kwargs["reciprocal"]
        )

        self._get_coordinates(coordinates)

    def __eq__(self, other):
        """Overrides the default implementation"""
        if hasattr(other, "subtype"):
            other = other.subtype
        if isinstance(other, MonotonicDimension):
            check = [
                self._count == other._count,
                np.all(self._coordinates == other._coordinates),
                self.reciprocal == other.reciprocal,
                super().__eq__(other),
            ]
            if np.all(check):
                return True
        return False

    def __repr__(self):
        properties = ", ".join(
            [
                f"coordinates={self._coordinates.__str__()}",
                *[
                    f"{k}={v}"
                    for k, v in self.dict().items()
                    if k not in ["type", "coordinates"]
                ],
            ]
        )
        return f"MonotonicDimension({properties})"

    def __str__(self):
        return f"MonotonicDimension({self.coordinates.__str__()})"

    def __mul__(self, other):
        """Multiply the MonotonicDimension object by a scalar."""
        return _update_monotonic_dimension_object_by_scalar(self.copy(), other, "mul")

    def __imul__(self, other):
        """Multiply the MonotonicDimension object by a scalar, in-place."""
        return _update_monotonic_dimension_object_by_scalar(self, other, "mul")

    def __truediv__(self, other):
        """Divide the MonotonicDimension object by a scalar."""
        return _update_monotonic_dimension_object_by_scalar(
            self.copy(), other, "truediv"
        )

    def __itruediv__(self, other):
        """Divide the MonotonicDimension object by a scalar, in-place."""
        return _update_monotonic_dimension_object_by_scalar(self, other, "truediv")

    def _get_coordinates(self, values):
        _unit = self._unit
        if isinstance(values, list):
            _coordinates = [
                ScalarQuantity(item, _unit).quantity.to(_unit).value for item in values
            ]
            self._coordinates = np.asarray(_coordinates, dtype=np.float64) * _unit
            self._values = values
            self._count = self._coordinates.size
            return

        if isinstance(values, Quantity):
            self._coordinates = values
        elif isinstance(values, np.ndarray):
            self._coordinates = values * _unit

        unit = scalar_quantity_format(self._coordinates[0], numerical_value=False)
        self._values = [f"{item.value} {unit}" for item in self._coordinates]
        self._count = self._coordinates.size

    # ----------------------------------------------------------------------- #
    #                                 Attributes                              #
    # ----------------------------------------------------------------------- #
    @property
    def type(self):
        """Return the type of the dimension."""
        return self.__class__._type

    @property
    def count(self):
        r"""Total number of points along the monotonic dimension."""
        return self._count

    @count.setter
    def count(self, value):
        self._count = check_count(value, self._count, "monotonic")

    @property
    def coordinates_offset(self):
        raise AttributeError(attribute_error(self, "coordinates_offset"))

    @property
    def coordinates(self):
        """Return the coordinates along the dimensions."""
        n = self._count
        coordinates = self._coordinates[:n]

        equivalent_fn = self._equivalencies

        if equivalent_fn is None:
            return coordinates.to(self._unit)

        equivalent_unit = self._equivalent_unit
        if equivalent_fn == "nmr_frequency_ratio":
            denominator = self.origin_offset - coordinates[0]
            if denominator.value == 0:
                raise ZeroDivisionError("Cannot convert the coordinates to ppm.")
            return coordinates.to(equivalent_unit, frequency_ratio(denominator))

        return coordinates.to(equivalent_unit, equivalent_fn)

    @coordinates.setter
    def coordinates(self, value):
        self._get_coordinates(value)

    @property
    def coords(self):
        """Alias for the `coordinates` attribute."""
        return self.coordinates

    @coords.setter
    def coords(self, value):
        self.coordinates = value

    @property
    def absolute_coordinates(self):
        """Return the absolute coordinates along the dimensions."""
        return (self.coordinates + self.origin_offset).to(self._unit)

    @property
    def axis_label(self):
        """Return a formatted string for displaying label along the dimension axis."""
        label = self.quantity_name if self.label.strip() == "" else self.label
        unit = (
            self._equivalent_unit if self._equivalent_unit is not None else self._unit
        )
        return _axis_label(label, unit)

    @property
    def data_structure(self):
        """Json serialized string describing the MonotonicDimension class instance."""
        return json.dumps(self.dict(), ensure_ascii=False, sort_keys=False, indent=2)

    # ----------------------------------------------------------------------- #
    #                                 Methods                                 #
    # ----------------------------------------------------------------------- #

    def _copy_metadata(self, obj, copy=False):
        """Copy MonotonicDimension metadata."""
        if hasattr(obj, "subtype"):
            obj = obj.subtype
        if isinstance(obj, MonotonicDimension):
            _copy_core_metadata(self, obj, "monotonic")

    def to_dict(self):
        """Alias to the `dict()` method of the class."""
        return self.dict()

    def dict(self):
        """Return the MonotonicDimension as a python dictionary."""
        dictionary = {}

        dictionary["type"] = self.__class__._type

        if self._description.strip() != "":
            dictionary["description"] = self._description.strip()

        dictionary["coordinates"] = self._values
        dictionary.update(self._dict())

        reciprocal_dictionary = {}
        if self.reciprocal._description.strip() != "":
            reciprocal_dictionary["description"] = self.reciprocal._description.strip()
        reciprocal_dictionary.update(self.reciprocal._dict())
        if reciprocal_dictionary == {}:
            del reciprocal_dictionary
        else:
            dictionary["reciprocal"] = reciprocal_dictionary

        return dictionary

    def copy(self):
        """Return a copy of the object."""
        return deepcopy(self)


def _update_monotonic_dimension_object_by_scalar(object_, other, type_):
    """Update object by multiplying by a scalar."""
    other = check_scalar_object(other)

    if type_ == "mul":
        object_._coordinates *= other
        object_._coordinates_offset *= other
        object_._origin_offset *= other
        object_._period *= other

    if type_ == "truediv":
        object_._coordinates /= other
        object_._coordinates_offset /= other
        object_._origin_offset /= other
        object_._period /= other

    object_._values = [str(item) for item in object_._coordinates]
    object_._unit = object_._coordinates.unit
    object_._quantity_name = object_._unit.physical_type
    object_._equivalencies = None
    _reciprocal_unit = object_._unit ** -1
    object_.reciprocal = ReciprocalDimension(unit=_reciprocal_unit)
    return object_



# -*- coding: utf-8 -*-
"""
Unit library parse extension.

The methods here parse a string containing physical
quantities to numerical value and unit. The unit
manipulation and conversion is done using
Astropy unit library.
"""
import warnings

import numexpr as ne
from astropy import units as u
from astropy.units import cds
from astropy.units import Quantity
from numpy import inf

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["string_to_quantity", "ScalarQuantity"]

NUMERIC = "0123456789-+.eE*/j^ ()"

tr = u.def_unit(["tr", "turn", "revolution"], 1.0 * u.cycle)
u.add_enabled_units([cds.ppm, tr])

convert = {
    "Å": "Angstrom",
    "°C": "deg_C",
    "°F": "deg_F",
    "°": "deg",
    "µ": "u",
    "ℏ": "/h",
    "Ω": "Ohm",
}


def get_start_index_of_quantity_unit(string):
    """Return the index where the unit of the quantity starts."""
    string = string.strip() + " "

    for i, c in enumerate(string):
        if c not in NUMERIC:
            break

    j = 1
    for j in range(1, i + 1):
        if string[:i][-j] == "(":
            break
        if string[:i][-j] == ")":
            j = j - 1
            break

    index = i if i == j else i - j
    return index


def string_to_quantity(string, dtype=float):
    """
    Parse the string and return a ``Quantity`` object.

    The string must contain a physical quantity. The method parse the string
    for numerical value and unit.

    :returns: ``Quantity`` object
    """
    index = get_start_index_of_quantity_unit(string)

    # the string has numerical value and unit
    if index != -1:
        try:
            number = float(ne.evaluate(string[:index]))
        except ZeroDivisionError:
            number = inf
        except Exception as e:
            raise ValueError(e)
    # the string is unit only
    else:
        index = 0
        number = 1.0

    unit = string[index:].strip()
    if unit != "" and unit != "()":
        if unit[0] == "(" and unit[-1] == ")":
            unit = unit[1:-1]

    for key in convert:
        unit = unit.replace(key, convert[key])
        unit_multiplier = 1
    try:
        unit_qt = u.Unit(unit) * unit_multiplier
        analysis = dtype(number) * unit_qt
        return analysis
    except BaseException as e:
        raise BaseException(e)


def scalar_quantity_format(quantity, numerical_value=True):
    """Convert unit to value object."""
    string = quantity.unit.to_string("fits").strip()
    for key in convert.keys():
        string = string.replace(convert[key], key)
    lst = {"10**-6": "ppm", "10**-2": "%"}
    for key in lst.keys():
        string = string.replace(key, lst[key])

    if numerical_value:
        cat_string = [str(quantity.value)]
    else:
        cat_string = [""]
    subunits = string.split(" ")
    for item in subunits:
        power = False
        if item.find("-") == -1:
            for i, c in enumerate(item):
                if c.isnumeric():
                    if i == 0:
                        break
                    cat_string.append(item[:i] + "^" + item[i:] + " *")
                    power = True
                    break
            if not power:
                cat_string.append(item + " *")
        else:
            l, r = item.split("-")
            cat_string.append(l + "^-" + r + " *")
    string = " ".join(cat_string)[:-2]
    string = string.replace("* / *", "/")
    return string.strip()


# def unit_to_latex(unit):
#     """NotImplemented."""
#     string = unit.to_string("fits").strip()
#     convert_tex = {
#         "Angstrom": "\\AA",
#         "deg_C": "$^\\circ$C",
#         "deg_F": "$^\\circ$C",
#         "deg": "$^\\circ$",
#         "u": "$\\mu$",
#         "/h": "$\\hbar$",
#         "Ohm": "$\\Ohm$",
#         "10**-6": "ppm",
#     }
#     for key in convert_tex:
#         string = string.replace(key, convert_tex[key])
#     lst = {"10**-6": "ppm", "10**-2": "%"}
#     for key in lst.keys():
#         string = string.replace(key, lst[key])

#     cat_string = []
#     subunits = string.split(" ")
#     for item in subunits:
#         power = False
#         if item.find("-") == -1:
#             for i, c in enumerate(item):
#                 if c.isnumeric():
#                     if i == 0:
#                         break
#                     cat_string.append(item[:i] + "$^{" + item[i:] + "}$ ")
#                     power = True
#                     break
#             if not power:
#                 cat_string.append(item + " ")
#         else:
#             l, r = item.split("-")
#             cat_string.append(l + "$^{-" + r + "}$ ")
#     string = " ".join(cat_string)
#     return string


class ScalarQuantity:

    """
    A ScalarQuantity class compliant with the CSDM standards.

    @params: quantity: Return the quantity object from astropy.units
                       library.
    """

    __slots__ = "quantity"

    def __init__(self, quantity_string=None, unit=None):
        self.quantity = self.quantity_object(quantity_string, unit)

    @classmethod
    def quantity_object(cls, quantity_string, unit):
        if isinstance(quantity_string, ScalarQuantity):
            return quantity_string.quantity

        if isinstance(quantity_string, Quantity):
            return quantity_string

        lst = [None, ""]
        if quantity_string in lst and unit in lst:
            return 0.0 * u.Unit("")

        if quantity_string in lst and unit not in lst:
            return 0.0 * unit

        if isinstance(quantity_string, str):
            quantity = string_to_quantity(quantity_string)
            if unit is not None:
                return check_unit_consistency(quantity, unit)
            return quantity

    def __str__(self):
        return self.__format__()

    def __format__(self, format="quantity"):
        """Format the units according to csdmpy recommendation.

        :ivar: format: The value can either be 'quantity' or 'unit'

        .. doctest::

            >>> a = ScalarQuantity('5 kg m^2 /s')
            >>> print(a.quantity)
            5.0 kg m2 / s
            >>> print(str(ScalarQuantity(a.quantity)))
            5.0 kg * m^2 * s^-1
            >>> print(ScalarQuantity(a.quantity).__format__('unit'))
            kg * m^2 * s^-1
        """

        element = _default_units(self.quantity)
        if format == "unit":
            return scalar_quantity_format(element, numerical_value=False)
        if format == "quantity":
            return scalar_quantity_format(element)


def check_unit_consistency(element, unit):
    if isinstance(unit, str):
        unit = ScalarQuantity(unit).quantity.unit
    if element.unit.physical_type != unit.physical_type:
        options = [
            str(element.unit),
            str(element.unit.physical_type),
            str(unit),
            unit.physical_type,
        ]
        message = (
            "Validation Failed: The unit '{0}' ({1}) is inconsistent "
            "with the unit '{2}' ({3})."
        )
        raise Exception(message.format(*options))
    return element


def _default_units(element):
    if element.unit.physical_type == "frequency":
        element = element.to("Hz")
    return element


def check_quantity_name(element, unit):
    if element is None:
        element = unit.physical_type
        return element

    if unit.physical_type == "unknown":
        warnings.warn(
            f"The physical quantity name associated with the unit, {str(unit)}, is not"
            f" defined in astropy.units package. Continuing with '{element}' as the "
            f"physical quantity name."
        )
        return element

    if element.lower() != unit.physical_type:
        warnings.warn(
            f"The physical quantity name, '{element}', is not defined in the "
            f"astropy.units package. Continuing with '{element}' as the physical "
            f"quantity name for unit {str(unit)}."
        )

    return element.lower()


def frequency_ratio(offset):
    return [
        (u.Hz, u.cds.ppm, lambda x: (x / offset.to("Hz")) * 1e6, lambda x: x * offset)
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from setuptools import find_packages
from setuptools import setup

# Load package veersion number.
with open("csdmpy/__init__.py", "r") as f:
    for line in f.readlines():
        if "__version__" in line:
            before_keyword, keyword, after_keyword = line.partition("=")
            version = after_keyword.strip()[1:-1]


# What packages are required for this module to be executed?
required = [
    "numpy>=1.17",
    "setuptools>=27.3",
    "astropy>=3.0",
    "requests>=2.21.0",
]
extras = {"matplotlib": ["matplotlib>=3.0"]}
setup_requires = ["setuptools>=27.3"]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = ""


setup(
    name="csdmpy",
    version=version,
    description="A python module for the core scientific dataset model.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Deepansh Srivastava",
    author_email="srivastava.89@osu.edu",
    python_requires=">=3.6",
    url="https://github.com/DeepanshS/csdmpy/",
    packages=find_packages(),
    install_requires=required,
    setup_requires=setup_requires,
    extras_require=extras,
    tests_require=["pytest", "pytest-runner"],
    include_package_data=True,
    license="BSD-3-Clause",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
)
